# -*- coding: utf-8 -*-

from copy import copy
from pandas import Grouper
from geopandas import GeoDataFrame
import numpy as np
import warnings

from .trajectory_stop_detector import TrajectoryStopDetector
from .trajectory import Trajectory
from .trajectory_collection import TrajectoryCollection
from .trajectory_utils import convert_time_ranges_to_segments
from .spatiotemporal_utils import TRange
from .geometry_utils import angular_difference


class TrajectorySplitter:
    """
    Splitter base class
    """

    def __init__(self, traj):
        """
        Create TrajectoryGeneralizer

        Parameters
        ----------
        traj : Trajectory or TrajectoryCollection
        """
        self.traj = traj

    def split(self, **kwargs):
        """
        Split the input Trajectory/TrajectoryCollection.

        Parameters
        ----------
        kwargs : any type
            Split parameters, differs by splitter

        Returns
        -------
        TrajectoryCollection
            Split trajectories
        """
        if isinstance(self.traj, Trajectory):
            return self._split_traj(self.traj, **kwargs)
        elif isinstance(self.traj, TrajectoryCollection):
            return self._split_traj_collection(**kwargs)
        else:
            raise TypeError

    def _split_traj_collection(self, **kwargs):
        trips = []
        for traj in self.traj:
            for x in self._split_traj(traj, **kwargs):
                if x.get_length() > self.traj.min_length:
                    trips.append(x)
        result = copy(self.traj)
        result.trajectories = trips
        return result

    def _split_traj(self, traj, **kwargs):
        return traj


class TemporalSplitter(TrajectorySplitter):
    """
    Split trajectories into subtrajectories using regular time intervals.

    Parameters
    ----------
    mode : str
        Split mode. (hour, day, month or year)
    min_length : numeric
        Desired minimum length of trajectories. (Shorter trajectories are discarded.)

    Examples
    --------

    >>> mpd.TemporalSplitter(traj).split(mode="year")
    """

    def _split_traj(self, traj, mode="day", min_length=0):
        result = []
        modes = {"hour": "h", "day": "D", "month": "ME", "year": "YE"}
        if mode in modes.keys():
            mode = modes[mode]
        grouped = traj.df.groupby(Grouper(freq=mode))
        dfs = []
        show_warning = False
        for _, values in grouped:
            if len(values) == 0:
                show_warning = True
            else:
                dfs.append(values)
        if show_warning:
            warnings.warn(
                f"Temporal splitting results contain observation gaps that exceed your "
                f"split size of {mode}. Consider running the ObservationGapSplitter to "
                f"further clean the results."
            )

        for i, df in enumerate(dfs):
            if i < len(dfs) - 1:
                next_index = dfs[i + 1].iloc[0].name
                next_values = dfs[i + 1].iloc[0].to_dict()
                df.loc[next_index] = next_values
                df = df.sort_index(ascending=True)
            if len(df) > 1:
                result.append(
                    Trajectory(df, f"{traj.id}_{i}", traj_id_col=traj.get_traj_id_col())
                )
        return TrajectoryCollection(result, min_length=min_length)


class ObservationGapSplitter(TrajectorySplitter):
    """
    Split trajectories into subtrajectories whenever there is a gap in the observations.

    Parameters
    ----------
    gap : datetime.timedelta
        Time gap threshold
    min_length : numeric
        Desired minimum length of trajectories. Shorter trajectories are discarded.
        (Length is calculated using CRS units, except if the CRS is geographic
        (e.g. EPSG:4326 WGS84) then length is calculated in metres.)

    Examples
    --------

    >>> mpd.ObservationGapSplitter(traj).split(gap=timedelta(hours=1))
    """

    def _split_traj(self, traj, gap, min_length=0):
        result = []
        temp_df = traj.df.copy()
        temp_df["t"] = temp_df.index
        temp_df["gap"] = temp_df["t"].diff() > gap
        temp_df["gap"] = temp_df["gap"].apply(lambda x: 1 if x else 0).cumsum()
        dfs = [group[1] for group in temp_df.groupby(temp_df["gap"])]
        for i, df in enumerate(dfs):
            df = df.drop(columns=["t", "gap"])
            if len(df) > 1:
                result.append(
                    Trajectory(df, f"{traj.id}_{i}", traj_id_col=traj.get_traj_id_col())
                )
        return TrajectoryCollection(result, min_length=min_length)


class SpeedSplitter(TrajectorySplitter):
    """
    Split trajectories if there are no speed measurements above the speed limit
    for the specified duration.

    Parameters
    ----------
    speed : float
        Speed limit
    duration : datetime.timedelta
        Minimum stop duration
    min_length : numeric
        Desired minimum length of trajectories. Shorter trajectories are discarded.
        (Length is calculated using CRS units, except if the CRS is geographic
        (e.g. EPSG:4326 WGS84) then length is calculated in metres.)
    max_speed: float
        Max speed limit
        (Speed is calculated as CRS units per second, except if the CRS is geographic
        (e.g. EPSG:4326 WGS84) then speed is calculated in meters per second.)

    Examples
    --------

    >>> mpd.SpeedSplitter(traj).split(speed=10, duration=timedelta(minutes=5))
    """

    def _split_traj(self, traj, speed, duration, min_length=0, max_speed=np.inf):
        traj = traj.copy()
        speed_col_name = traj.get_speed_col()
        if speed_col_name not in traj.df.columns:
            traj.add_speed(overwrite=True)
        traj.df = traj.df[traj.df[speed_col_name].between(speed, max_speed)]
        return ObservationGapSplitter(traj).split(gap=duration, min_length=min_length)


class StopSplitter(TrajectorySplitter):
    """
    Split trajectories at detected stops.
    A stop is detected if the movement stays within an area of specified size for
    at least the specified duration.

    Parameters
    ----------
    max_diameter : float
        Maximum diameter for stop detection
    min_duration : datetime.timedelta
        Minimum stop duration
    min_length : numeric
        Desired minimum length of trajectories. Shorter trajectories are discarded.
        (Length is calculated using CRS units, except if the CRS is geographic
        (e.g. EPSG:4326 WGS84) then length is calculated in metres.)

    Examples
    --------

    >>> mpd.StopSplitter(traj).split(max_diameter=7, min_duration=timedelta(seconds=60))
    """

    def _split_traj(self, traj, max_diameter, min_duration, min_length=0):
        stop_detector = TrajectoryStopDetector(traj)
        stop_time_ranges = stop_detector.get_stop_time_ranges(
            max_diameter, min_duration
        )
        between_stops = self.get_time_ranges_between_stops(traj, stop_time_ranges)
        result = convert_time_ranges_to_segments(traj, between_stops)
        return TrajectoryCollection(result, min_length=min_length)

    @staticmethod
    def get_time_ranges_between_stops(traj, stop_ranges):
        items = [traj.get_start_time()]
        for stop_range in stop_ranges:
            items.append(stop_range.t_0)
            items.append(stop_range.t_n)
        items.append(traj.get_end_time())
        result = [TRange(*items[x : x + 2]) for x in range(0, len(items), 2)]
        return result


class AngleChangeSplitter(TrajectorySplitter):
    """
    Split trajectories into subtrajectories whenever there is a specified change in
    the heading angle.

    Parameters
    ----------
    min_angle : float
        Minimum angle change
    min_speed: float
        Min speed threshold at which points are considered for angle change.
        (Speed is calculated as CRS units per second, except if the CRS is geographic
        (e.g. EPSG:4326 WGS84) then speed is calculated in meters per second.)
    min_length : numeric
        Desired minimum length of trajectories. Shorter trajectories are discarded.
        (Length is calculated using CRS units, except if the CRS is geographic
        (e.g. EPSG:4326 WGS84) then length is calculated in metres.)

    Examples
    --------

    >>> mpd.AngleSplitter(traj).split(min_angle=45, min_speed=15)
    """

    def _split_traj(self, traj, min_angle=45, min_speed=0, min_length=0):
        result = []
        traj = traj.copy()

        direction_col_name = traj.get_direction_col()
        if direction_col_name not in traj.df.columns:
            traj.add_direction(overwrite=True)

        speed_col_name = traj.get_speed_col()
        if speed_col_name not in traj.df.columns:
            traj.add_speed(overwrite=True)

        comp_dir = traj.df[direction_col_name].iloc[0]
        traj.df["dirChange"] = -1
        dir_group = 0

        for i, (direction, speed) in enumerate(
            zip(traj.df[direction_col_name].tolist(), traj.df[speed_col_name].tolist())
        ):
            if speed >= min_speed:
                if angular_difference(comp_dir, direction) >= min_angle:
                    comp_dir = direction
                    dir_group += 1

            traj.df.iloc[i, traj.df.columns.get_loc("dirChange")] = dir_group

        dfs = [group[1] for group in traj.df.groupby(traj.df["dirChange"])]
        for i, df in enumerate(dfs):
            df = df.drop(columns=["dirChange"])
            if len(df) > 1:
                prev_index = dfs[i - 1].iloc[-1].name
                prev_values = dfs[i - 1].iloc[-1].to_dict()
                if i > 0:
                    df.loc[prev_index] = prev_values
                    df = df.sort_index(ascending=True)

                result.append(
                    Trajectory(df, f"{traj.id}_{i}", traj_id_col=traj.get_traj_id_col())
                )

        return TrajectoryCollection(result, min_length=min_length)


class ValueChangeSplitter(TrajectorySplitter):
    """
    Split trajectories into subtrajectories whenever there is a change in
    the specified column values.

    Parameters
    ----------
    col_name : string
        Name of the col to monitor for changes in consecutive values
    min_length : numeric
        Desired minimum length of trajectories. Shorter trajectories are discarded.
        (Length is calculated using CRS units, except if the CRS is geographic
        (e.g. EPSG:4326 WGS84) then length is calculated in metres.)

    Examples
    --------

    >>> mpd.ValueChangeSplitter(traj).split(col_name='column1')
    """

    def _split_traj(self, traj, col_name, min_length=0):
        result = []
        temp_df = traj.df.copy()
        temp_df["t"] = temp_df.index
        temp_df["change"] = temp_df[col_name].shift() != temp_df[col_name]
        temp_df["change"] = temp_df["change"].apply(lambda x: 1 if x else 0).cumsum()
        dfs = [group[1] for group in temp_df.groupby(temp_df["change"])]
        for i, df in enumerate(dfs):
            df = df.drop(columns=["t", "change"])
            if i < (len(dfs) - 1):
                next_index = dfs[i + 1].iloc[0].name
                next_values = dfs[i + 1].iloc[0].to_dict()
                df.loc[next_index] = next_values
                df = df.sort_index(ascending=True)
            if len(df) > 1:
                df = GeoDataFrame(df).set_crs(traj.df.crs)
                traj = Trajectory(
                    df, f"{traj.id}_{i}", traj_id_col=traj.get_traj_id_col()
                )
                result.append(traj)
        return TrajectoryCollection(result, min_length=min_length)



# -*- coding: utf-8 -*-

import pandas as pd
from geopandas import GeoDataFrame
from shapely.geometry import Point
from pyproj import CRS
from datetime import datetime, timedelta
from .test_trajectory import make_traj, Node
from movingpandas.trajectory import Trajectory
from movingpandas.trajectory_collection import TrajectoryCollection
from movingpandas.trajectory_generalizer import (
    MaxDistanceGeneralizer,
    MinDistanceGeneralizer,
    MinTimeDeltaGeneralizer,
    DouglasPeuckerGeneralizer,
    TopDownTimeRatioGeneralizer,
)


CRS_METRIC = CRS.from_user_input(31256)
CRS_LATLON = CRS.from_user_input(4326)


class TestTrajectoryGeneralizer:
    def setup_method(self):
        self.nodes = [
            Node(0, 0, day=1),
            Node(1, 0.1, day=2),
            Node(2, 0.2, day=3),
            Node(3, 0, day=4),
            Node(3, 3, day=5),
        ]
        self.traj = make_traj(self.nodes)
        df = pd.DataFrame(
            [{"xxx": n.geometry, "t": n.t} for n in self.nodes]
        ).set_index("t")
        geo_df = GeoDataFrame(df, geometry="xxx", crs=CRS_METRIC)
        self.traj_other_geometry_column_names = Trajectory(geo_df, 1)
        df = pd.DataFrame(
            [
                [1, "A", Point(0, 0), datetime(2018, 1, 1, 12, 0, 0), 9, "a"],
                [1, "A", Point(6, 0), datetime(2018, 1, 1, 12, 6, 0), 5, "b"],
                [1, "A", Point(6, 6), datetime(2018, 1, 1, 14, 10, 0), 2, "c"],
                [1, "A", Point(9, 9), datetime(2018, 1, 1, 14, 15, 0), 4, "d"],
                [2, "A", Point(10, 10), datetime(2018, 1, 1, 12, 0, 0), 10, "e"],
                [2, "A", Point(16, 10), datetime(2018, 1, 1, 12, 6, 0), 6, "f"],
                [2, "A", Point(16, 16), datetime(2018, 1, 2, 13, 10, 0), 7, "g"],
                [2, "A", Point(190, 19), datetime(2018, 1, 2, 13, 15, 0), 3, "h"],
            ],
            columns=["id", "obj", "geometry", "t", "val", "val2"],
        ).set_index("t")
        self.geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        self.collection = TrajectoryCollection(self.geo_df, "id", obj_id_col="obj")

    def test_douglas_peucker(self):
        result = DouglasPeuckerGeneralizer(self.traj).generalize(tolerance=1)
        assert result == make_traj([self.nodes[0], self.nodes[3], self.nodes[4]])

    def test_douglas_peucker_for_other_geometry_column_names(self):
        result = DouglasPeuckerGeneralizer(
            self.traj_other_geometry_column_names
        ).generalize(tolerance=1)
        assert result == make_traj([self.nodes[0], self.nodes[3], self.nodes[4]])

    def test_tdtr(self):
        result = TopDownTimeRatioGeneralizer(self.traj).generalize(tolerance=1)
        assert result == make_traj([self.nodes[0], self.nodes[3], self.nodes[4]])

    def test_tdtr_different_than_dp(self):
        nodes = [
            Node(),
            Node(1, 0.1, hour=1),
            Node(1, 2, hour=7),
            Node(2, 2, hour=15),
            Node(3, 0, hour=16),
            Node(3, 3, hour=17),
        ]
        traj = make_traj(nodes)
        result = TopDownTimeRatioGeneralizer(traj).generalize(tolerance=1)
        assert result == make_traj([nodes[0], nodes[2], nodes[3], nodes[4], nodes[5]])

    def test_tdtr_for_other_geometry_column_names(self):
        result = TopDownTimeRatioGeneralizer(
            self.traj_other_geometry_column_names
        ).generalize(tolerance=1)
        assert result == make_traj([self.nodes[0], self.nodes[3], self.nodes[4]])

    def test_max_distance(self):
        result = MaxDistanceGeneralizer(self.traj).generalize(tolerance=1)
        assert result == make_traj([self.nodes[0], self.nodes[3], self.nodes[4]])

    def test_max_distance_for_other_geometry_column_names(self):
        result = MaxDistanceGeneralizer(
            self.traj_other_geometry_column_names
        ).generalize(tolerance=1)
        assert result == make_traj([self.nodes[0], self.nodes[3], self.nodes[4]])

    def test_min_time_delta(self):
        nodes = [
            Node(),
            Node(1, 0.1, minute=6),
            Node(2, 0.2, minute=10),
            Node(3, 0, minute=30),
            Node(3, 3, minute=59),
        ]
        traj = make_traj(nodes)
        result = MinTimeDeltaGeneralizer(traj).generalize(timedelta(minutes=10))
        assert result == make_traj([nodes[0], nodes[2], nodes[3], nodes[4]])

    def test_min_distance(self):
        nodes = [
            Node(),
            Node(0, 0.1, day=2),
            Node(0, 0.2, day=3),
            Node(0, 1, day=4),
            Node(0, 3, day=5),
        ]
        traj = make_traj(nodes, CRS_METRIC)
        result = MinDistanceGeneralizer(traj).generalize(tolerance=1)
        assert result == make_traj([nodes[0], nodes[3], nodes[4]], CRS_METRIC)

    def test_collection(self):
        collection = MinTimeDeltaGeneralizer(self.collection).generalize(
            tolerance=timedelta(minutes=10)
        )
        assert len(collection) == 2
        wkt1 = collection.trajectories[0].to_linestring().wkt
        assert wkt1 == "LINESTRING (0 0, 6 6, 9 9)"
        wkt2 = collection.trajectories[1].to_linestring().wkt
        assert wkt2 == "LINESTRING (10 10, 16 16, 190 19)"



"""
`movingpandas`: Implementation of Trajectory classes and functions built on top
of GeoPandas
"""

import warnings


from .trajectory import Trajectory  # noqa F401
from .trajectory_generalizer import (  # noqa F401
    TrajectoryGeneralizer,
    MaxDistanceGeneralizer,
    MinDistanceGeneralizer,
    MinTimeDeltaGeneralizer,
    DouglasPeuckerGeneralizer,
    TopDownTimeRatioGeneralizer,
)
from .trajectory_collection import TrajectoryCollection  # noqa F401
from .io import read_mf_json, read_mf_dict, gdf_to_mf_json  # noqa F401
from .trajectory_aggregator import (  # noqa F401
    TrajectoryCollectionAggregator,
    PtsExtractor,
)
from .trajectory_splitter import (  # noqa F401
    TrajectorySplitter,
    TemporalSplitter,
    ObservationGapSplitter,
    SpeedSplitter,
    StopSplitter,
    AngleChangeSplitter,
    ValueChangeSplitter,
)
from .trajectory_cleaner import IqrCleaner, OutlierCleaner  # noqa F401
from .trajectory_stop_detector import TrajectoryStopDetector  # noqa F401
from .point_clusterer import PointClusterer  # noqa F401
from .tools._show_versions import show_versions  # noqa F401

try:
    from .trajectory_smoother import KalmanSmootherCV  # noqa F401
except ImportError as e:
    warnings.warn(e.msg, UserWarning)

name = "movingpandas"
__version__ = "0.21.3"



# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt


class _TrajectoryPlotter:
    def __init__(self, data, *args, **kwargs):
        self.data = data
        self.args = args
        self.kwargs = kwargs

        self.width = kwargs.pop("width", 900)
        self.height = kwargs.pop("height", 700)
        self.figsize = kwargs.pop("figsize", None)
        self.column = kwargs.get("column", None)
        self.column = kwargs.get("c", self.column)
        self.ax = kwargs.pop("ax", None)
        self.colormap = kwargs.pop("colormap", None)
        self.colormap = kwargs.pop("column_to_color", self.colormap)
        self.clim = self.kwargs.pop("clim", None)
        self.colorbar = self.kwargs.pop("colorbar", True)

        self.min_value = kwargs.get("vmin", None)
        self.max_value = kwargs.get("vmax", None)

        self.overlay = None
        self.hvplot_is_geo = kwargs.pop("geo", True)
        self.hvplot_tiles = kwargs.pop("tiles", "OSM")

        self.plot_lines = kwargs.pop("plot_lines", True)
        self.marker_size = kwargs.pop("marker_size", 200)
        self.marker_color = kwargs.pop("marker_color", None)
        self.line_width = kwargs.pop("line_width", 3.0)

        self.column_names = data.get_column_names()
        self.traj_id_col_name = data.get_traj_id_col()
        self.speed_col_name = data.get_speed_col()
        self.direction_col_name = data.get_direction_col()
        self.geom_col_name = data.get_geom_col()
        self.speed_col_missing = self.speed_col_name not in self.column_names

        self.hv_defaults = {
            "width": self.width,
            "height": self.height,
            "active_tools": ["wheel_zoom"],
        }

    def get_min_max_values(self):
        min_value = self.data.get_min(self.column)
        max_value = self.data.get_max(self.column)
        self.min_value = self.kwargs.pop("vmin", min_value)
        self.max_value = self.kwargs.pop("vmax", max_value)
        return min_value, max_value

    def plot(self):
        if not self.ax:
            self.ax = plt.figure(figsize=self.figsize).add_subplot(1, 1, 1)
        tc = self.preprocess_data()
        line_plot = self._plot_lines(tc)

        to_drop = [x for x in tc.get_column_names() if x not in self.column_names]
        tc.drop(columns=to_drop)

        return line_plot

    def explore(self):
        tc = self.preprocess_data()
        explore_plot = self._explore_lines(tc)

        to_drop = [x for x in tc.get_column_names() if x not in self.column_names]
        tc.drop(columns=to_drop)

        return explore_plot

    def _explore_lines(self, tc):
        from importlib.metadata import version

        cols = [self.traj_id_col_name, self.geom_col_name]
        if self.column:
            cols = cols + [self.column]
        cols = list(set(cols))

        if self.column is None:
            line_gdf = tc.to_line_gdf()
        else:
            line_gdf = tc.to_line_gdf(columns=cols)

        if version("geopandas") >= "1.0.0":
            return line_gdf.explore(*self.args, **self.kwargs)
        else:
            raise (
                NotImplementedError(
                    "Please install geopandas >= 1.0.0 to use this function."
                )
            )

    def preprocess_data(self):
        tc = self.data

        if self.column is None:
            return tc

        if self.column:
            if self.column == self.speed_col_name and self.speed_col_missing:
                tc.add_speed()

            (min_val, max_val) = self.get_min_max_values()
            if self.clim is None:
                self.clim = (min_val, max_val)

        return tc

    def _plot_lines(self, tc):
        line_gdf = tc.to_line_gdf()

        if self.column and self.colormap:
            line_gdf["color"] = line_gdf[self.column].apply(
                lambda x: self.colormap[x] if x in self.colormap else "grey"
            )
            self.kwargs.pop(
                "column"
            )  # to avoid gpd warning (only specify one of 'column' or 'color')
            return line_gdf.plot(
                ax=self.ax, color=line_gdf["color"], *self.args, **self.kwargs
            )
        else:
            self.kwargs.pop("vmin", None)
            self.kwargs.pop("vmax", None)
            return line_gdf.plot(
                ax=self.ax,
                vmin=self.min_value,
                vmax=self.max_value,
                *self.args,
                **self.kwargs
            )

    def hvplot(self):  # noqa F811
        try:
            import hvplot.pandas  # noqa F401, seems necessary for the following import to work
            import colorcet as cc
            from holoviews import opts
            from bokeh.palettes import Category10_10
        except ImportError as error:
            raise ImportError(
                "Missing optional dependencies. To use interactive plotting, "
                "install hvplot and GeoViews (see "
                "https://hvplot.holoviz.org/getting_started/installation.html and "
                "https://geoviews.org)."
            ) from error

        opts.defaults(opts.Overlay(**self.hv_defaults))
        self.MPD_PALETTE = list(Category10_10) + cc.palette["glasbey"]

        self.color = self.kwargs.pop("color", None)

        tc = self.preprocess_data()

        plot = self._hvplot_lines(tc)
        if self.marker_size > 0:
            plot = plot * self._hvplot_end_points(tc)

        to_drop = [x for x in tc.get_column_names() if x not in self.column_names]
        tc.drop(columns=to_drop)

        if self.overlay:
            return self.overlay * plot
        else:
            return plot

    def _hvplot_end_points(self, tc):
        from holoviews import dim, Overlay

        if "TrajectoryCollection" in str(type(tc)):
            end_pts = tc.get_end_locations(with_direction=True)
        else:  # Trajectory
            tc.add_direction(name=self.direction_col_name, overwrite=True)
            end_pts = tc.df.tail(1).copy()

        end_pts["triangle_angle"] = end_pts[self.direction_col_name] * -1.0
        end_pts["triangle_angle"] = end_pts["triangle_angle"].astype(float)

        hover_cols = self.kwargs.pop("hover_cols", None)

        self.kwargs["hover_cols"] = ["triangle_angle"]
        if hover_cols:
            self.kwargs["hover_cols"] = self.kwargs["hover_cols"] + hover_cols
        if self.marker_color:
            self.kwargs["color"] = self.marker_color
        if self.column:
            self.kwargs["hover_cols"] = self.kwargs["hover_cols"] + [self.column]

        if self.hvplot_is_geo and not tc.is_latlon and tc.get_crs() is not None:
            end_pts = end_pts.to_crs(epsg=4326)

        if self.column:
            return end_pts.hvplot(
                geo=self.hvplot_is_geo,
                tiles=None,
                marker="triangle",
                angle=dim("triangle_angle"),
                size=self.marker_size,
                clim=self.clim,
                *self.args,
                **self.kwargs
            )
        else:
            plots = []
            for i in range(len(end_pts)):
                tmp = end_pts.iloc[i : i + 1, :]
                tmp = tmp.hvplot(
                    geo=self.hvplot_is_geo,
                    tiles=None,
                    marker="triangle",
                    angle=dim("triangle_angle"),
                    size=self.marker_size,
                    color=self.get_color(i),
                    *self.args,
                    **self.kwargs
                )
                plots.append(tmp)
            return Overlay(plots)

    def _hvplot_lines(self, tc):
        cols = [self.traj_id_col_name, self.geom_col_name]
        if "hover_cols" in self.kwargs:
            cols = cols + self.kwargs["hover_cols"]
        if self.column:
            cols = cols + [self.column]
        cols = list(set(cols))

        if self.column is None:
            return self._hvplot_traj_gdf(tc)
        else:
            return self._hvplot_line_gdf(tc, cols)

    def get_color(self, i):
        if self.color:
            return self.color
        else:
            return self.MPD_PALETTE[i]

    def _hvplot_traj_gdf(self, tc):
        from holoviews import Cycle, Overlay

        Cycle.default_cycles["default_colors"] = self.MPD_PALETTE

        traj_gdf = tc.to_traj_gdf()
        if self.hvplot_is_geo and not tc.is_latlon and tc.get_crs() is not None:
            traj_gdf = traj_gdf.to_crs(epsg=4326)

        plots = []
        for i in range(len(traj_gdf)):
            tmp = traj_gdf.iloc[i : i + 1, :]
            tmp = tmp.hvplot(
                line_width=self.line_width,
                geo=self.hvplot_is_geo,
                tiles=self.hvplot_tiles,
                color=self.get_color(i),
                colorbar=self.colorbar,
                *self.args,
                **self.kwargs
            )
            self.hvplot_tiles = None
            plots.append(tmp)
        return Overlay(plots)

    def _hvplot_line_gdf(self, tc, cols):
        line_gdf = tc.to_line_gdf(columns=cols)

        ids = None
        if self.column is None and self.traj_id_col_name is not None:
            ids = line_gdf[self.traj_id_col_name].unique()
        self.set_default_cmaps(ids)

        if self.hvplot_is_geo and not tc.is_latlon and tc.get_crs() is not None:
            line_gdf = line_gdf.to_crs(epsg=4326)

        return line_gdf.hvplot(
            line_width=self.line_width,
            geo=self.hvplot_is_geo,
            tiles=self.hvplot_tiles,
            clim=self.clim,
            colorbar=self.colorbar,
            *self.args,
            **self.kwargs
        )

    def hvplot_pts(self):
        try:
            import hvplot.pandas  # noqa F401, seems necessary for the following import to work
            import colorcet as cc
            from holoviews import opts, dim, Overlay
            from bokeh.palettes import Category10_10
        except ImportError as error:
            raise ImportError(
                "Missing optional dependencies. To use interactive plotting, "
                "install hvplot and GeoViews (see "
                "https://hvplot.holoviz.org/getting_started/installation.html and "
                "https://geoviews.org)."
            ) from error

        opts.defaults(opts.Overlay(**self.hv_defaults))
        self.MPD_PALETTE = list(Category10_10) + cc.palette["glasbey"]
        self.color = self.kwargs.pop("color", None)

        if "TrajectoryCollection" in str(type(self.data)):
            tc = self.data.copy()
            if self.direction_col_name not in tc.trajectories[0].df.columns:
                tc.add_direction(name=self.direction_col_name)
            if self.column:
                if self.column == self.speed_col_name and self.speed_col_missing:
                    tc.add_speed()
            pts_gdf = tc.to_point_gdf()
        else:  # Trajectory
            traj = self.data.copy()
            if self.direction_col_name not in traj.df.columns:
                traj.add_direction(name=self.direction_col_name)
            if self.column:
                if self.column == self.speed_col_name and self.speed_col_missing:
                    traj.add_speed()
            pts_gdf = traj.df

        ids = None
        if self.column is None and self.traj_id_col_name is not None:
            ids = pts_gdf[self.traj_id_col_name].unique()
        self.set_default_cmaps(ids)

        pts_gdf["triangle_angle"] = pts_gdf[self.direction_col_name] * -1.0
        pts_gdf["triangle_angle"] = pts_gdf["triangle_angle"].astype(float)
        pts_gdf["dash_angle"] = ((pts_gdf[self.direction_col_name] * -1.0) + 90).astype(
            float
        )

        hover_cols = self.kwargs.pop("hover_cols", None)

        self.kwargs["hover_cols"] = ["triangle_angle", "dash_angle"]
        if hover_cols:
            self.kwargs["hover_cols"] = self.kwargs["hover_cols"] + hover_cols
        if self.marker_color:
            self.kwargs["color"] = self.marker_color
        if self.column:
            self.kwargs["hover_cols"] = self.kwargs["hover_cols"] + [self.column]

        if (
            self.hvplot_is_geo
            and not self.data.is_latlon
            and self.data.get_crs() is not None
        ):
            pts_gdf = pts_gdf.to_crs(epsg=4326)

        if self.column:
            arrow_shaft = pts_gdf.hvplot(
                geo=self.hvplot_is_geo,
                tiles=None,
                marker="dash",
                angle=dim("dash_angle"),
                size=self.marker_size * 1.7,
                line_width=self.marker_size / 70.0,
                clim=self.clim,
                *self.args,
                **self.kwargs
            )
            arrow_head = pts_gdf.hvplot(
                geo=self.hvplot_is_geo,
                tiles=None,
                marker="triangle",
                angle=dim("triangle_angle"),
                size=self.marker_size,
                clim=self.clim,
                *self.args,
                **self.kwargs
            )
            return arrow_shaft * arrow_head
        else:
            plots = []
            for i in range(len(ids)):
                tmp = pts_gdf[pts_gdf[self.traj_id_col_name] == ids[i]]
                arrow_shaft = tmp.hvplot(
                    geo=self.hvplot_is_geo,
                    tiles=None,
                    marker="dash",
                    angle=dim("dash_angle"),
                    size=self.marker_size * 1.7,
                    line_width=self.marker_size / 70.0,
                    color=self.get_color(i),
                    *self.args,
                    **self.kwargs
                )
                arrow_head = tmp.hvplot(
                    geo=self.hvplot_is_geo,
                    tiles=None,
                    marker="triangle",
                    angle=dim("triangle_angle"),
                    size=self.marker_size,
                    color=self.get_color(i),
                    *self.args,
                    **self.kwargs
                )
                plots.append(arrow_shaft * arrow_head)
            return Overlay(plots)

    def set_default_cmaps(self, ids=None):
        from holoviews import Cycle

        Cycle.default_cycles["default_colors"] = self.MPD_PALETTE
        if self.column == self.speed_col_name and "cmap" not in self.kwargs:
            self.kwargs["cmap"] = "Plasma"
        elif self.column is None and self.traj_id_col_name is not None:
            self.kwargs["c"] = self.traj_id_col_name
            if "cmap" not in self.kwargs:
                print("building colormap ...")
                self.kwargs["colormap"] = dict(zip(ids, self.MPD_PALETTE[: len(ids)]))
                print(self.kwargs["colormap"])
        if self.colormap:
            self.kwargs["colormap"] = self.colormap
        if "cmap" not in self.kwargs and "colormap" not in self.kwargs:
            self.kwargs["cmap"] = self.MPD_PALETTE



# -*- coding: utf-8 -*-
import json
import os

import pandas as pd
import pytest

from movingpandas.io import (
    _create_objects_from_mf_json_dict,
    gdf_to_mf_json,
    read_mf_json,
    read_mf_dict,
)


class TestIO:
    test_dir = os.path.dirname(os.path.realpath(__file__))

    def test_mf_file_movingpoint(self):
        traj = read_mf_json(os.path.join(self.test_dir, "movingfeatures.json"), "id")
        assert traj.id == "9569"
        assert traj.size() == 5
        actual = traj.df.columns
        expected = ["pressure", "wind", "class", "geometry", "id"]
        assert len(actual) == len(expected)
        assert all([a == b for a, b in zip(actual, expected)])
        actual = list(traj.df["pressure"])
        expected = [1004.0, 1004.0, 1004.0, 1004.0, 1000.0]
        assert actual == expected

    def test_mf_file_mftrajectory(self):
        traj = read_mf_json(
            os.path.join(self.test_dir, "mftrajectory.json"), traj_id=9569
        )
        assert traj.id == 9569
        assert traj.size() == 5
        actual = traj.df.columns
        expected = ["pressure", "wind", "class", "geometry", "traj_id"]
        assert len(actual) == len(expected)
        assert all([a == b for a, b in zip(actual, expected)])
        actual = list(traj.df["pressure"])
        expected = [1004.0, 1004.0, 1004.0, 1004.0, 1000.0]
        assert actual == expected

    def test_read_mf_dict(self):
        collection = {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "properties": {"id": 5},
                    "temporalGeometry": {
                        "type": "MovingPoint",
                        "datetimes": ["2008-02-02T15:02:18Z", "2008-02-02T18:32:28Z"],
                        "coordinates": [[116.52299, 40.07757], [116.52302, 39.92129]],
                    },
                }
            ],
        }
        trajs_collection = read_mf_dict(collection, traj_id_property="id")

        assert len(trajs_collection) == 1
        assert trajs_collection.get_trajectory(5).id == 5

        traj = read_mf_dict(collection["features"][0], traj_id_property="id")

        assert traj.id == 5

    def test_mf_collection_file(self):
        trajs_collection = read_mf_json(
            os.path.join(self.test_dir, "movingfeatures_collection.json"), "id"
        )
        assert len(trajs_collection) == 2
        assert trajs_collection.get_trajectory(1).id == 1
        assert trajs_collection.get_trajectory(2).id == 2
        actual = list(trajs_collection.get_trajectory(2).df["pressure"])[:5]
        expected = [
            1008.0,
            1006.0,
            1006.0,
            1006.0,
            1006.0,
        ]
        assert actual == expected

    def test_mf_collection_needs_id(self):
        with pytest.raises(AssertionError):
            read_mf_json(
                os.path.join(self.test_dir, "movingfeatures_collection.json"), None
            )

    def test_wrong_property_raises_error(self):
        with pytest.raises(ValueError):
            read_mf_json(os.path.join(self.test_dir, "movingfeatures.json"), "foo")

    def test_dict(self):
        data = {
            "type": "Feature",
            "properties": {"id": 5},
            "temporalGeometry": {
                "type": "MovingPoint",
                "datetimes": ["2008-02-02T15:02:18Z", "2008-02-02T18:32:28Z"],
                "coordinates": [[116.52299, 40.07757], [116.52302, 39.92129]],
            },
        }
        traj = _create_objects_from_mf_json_dict(data, "id")
        assert traj.id == 5
        assert traj.size() == 2

    def test_dict_without_id(self):
        data = {
            "type": "Feature",
            "temporalGeometry": {
                "type": "MovingPoint",
                "datetimes": ["2008-02-02T15:02:18Z", "2008-02-02T18:32:28Z"],
                "coordinates": [[116.52299, 40.07757], [116.52302, 39.92129]],
            },
        }
        traj = _create_objects_from_mf_json_dict(data)
        assert traj.id == 0
        assert traj.size() == 2

    def test_no_dict(self):
        data = []
        with pytest.raises(ValueError):
            _create_objects_from_mf_json_dict(data, "id")

    def test_invalid_feature_collection(self):
        data = {"type": "FeatureCollection", "temporalGeometry": []}
        with pytest.raises(ValueError):
            _create_objects_from_mf_json_dict(data, "id")

    def test_unsupported_geometry_type(self):
        data = {"type": "Feature", "temporalGeometry": {"type": "MovingPolygon"}}
        with pytest.raises(RuntimeError):
            _create_objects_from_mf_json_dict(data, "id")

    def test_gdf_to_mf_json(self):
        # Load a GeoDataFrame from a Moving-Features JSON file.
        loaded_gdf = read_mf_json(
            os.path.join(self.test_dir, "movingfeatures.json"), "id"
        ).df

        loaded_gdf["t"] = loaded_gdf.index
        loaded_gdf["id"] = "9569"

        # Convert the GeoDataFrame to a Moving-Features JSON dictionary.
        entity_mf_json = json.loads(
            json.dumps(
                gdf_to_mf_json(
                    loaded_gdf,
                    traj_id_column="id",
                    datetime_column="t",
                    datetime_encoder=lambda x: x.isoformat() + "Z",
                    temporal_columns=["wind", "pressure", "class"],
                    interpolation="Linear",
                    temporal_columns_static_fields={
                        "wind": {
                            "form": "KNT",
                            "interpolation": "Linear",
                            "type": "Measure",
                        },
                        "pressure": {
                            "form": "A97",
                            "interpolation": "Linear",
                            "type": "Measure",
                        },
                        "class": {"interpolation": "Linear", "type": "Measure"},
                    },
                )["features"][0]
            )
        )

        # Load the expected result from the original Moving-Features JSON file.
        with open(os.path.join(self.test_dir, "movingfeatures.json"), "r") as f:
            expected_mf_json = json.load(f)

        # Compare the expected and actual Moving-Features JSON dictionaries.
        assert entity_mf_json == expected_mf_json

    def test_not_geodataframe_raises_error(self):
        with pytest.raises(TypeError):
            gdf_to_mf_json(
                pd.DataFrame(),
                traj_id_column="id",
                datetime_column="t",
            )

    def test_missing_datetime_column_raises_error(self):
        with pytest.raises(TypeError):
            gdf_to_mf_json(
                pd.DataFrame(),
                traj_id_column="id",
            )

    def test_missing_traj_id_property_raises_error(self):
        with pytest.raises(TypeError):
            gdf_to_mf_json(
                pd.DataFrame(),
                datetime_column="t",
            )



import pandas as pd
from geopandas import GeoDataFrame
from shapely.geometry import Point

from movingpandas import PointClusterer


class TestClustering:
    def test_cluster_points(self):
        pts = [Point(0, 0), Point(10, 10), Point(10.2, 10.2)]
        expected = GeoDataFrame(
            pd.DataFrame(
                [
                    {"geometry": Point(0, 0), "n": 1},
                    {"geometry": Point(10.1, 10.1), "n": 2},
                ]
            )
        )
        actual = PointClusterer(pts, max_distance=3, is_latlon=False).get_clusters()
        actual = [(c.centroid, len(c.points)) for c in actual]
        expected = [(c.geometry, c.n) for key, c in expected.iterrows()]
        assert len(actual) == len(expected)
        for pt in expected:
            assert pt in actual

    def test_cluster_points2(self):
        pts = [
            Point(0, 0),
            Point(6, 0),
            Point(6, 6),
            Point(0.2, 0.2),
            Point(6.2, 0.2),
            Point(6.2, 6.2),
        ]
        expected = GeoDataFrame(
            pd.DataFrame(
                [
                    {"geometry": Point(0.1, 0.1), "n": 2},
                    {"geometry": Point(6.1, 0.1), "n": 2},
                    {"geometry": Point(6.1, 6.1), "n": 2},
                ]
            )
        )
        actual = PointClusterer(pts, max_distance=5, is_latlon=False).get_clusters()
        actual = [(c.centroid, len(c.points)) for c in actual]
        expected = [(c.geometry, c.n) for key, c in expected.iterrows()]
        assert len(actual) == len(expected)
        for pt in expected:
            assert pt in actual

    def test_cluster_horizontal_points(self):
        pts = [
            Point(0, 2),
            Point(0, 6),
            Point(0, 4),
            Point(0, 3),
            Point(0, 5),
            Point(0, 7),
        ]
        expected = GeoDataFrame(pd.DataFrame([{"geometry": Point(0, 4.5), "n": 6}]))
        actual = PointClusterer(pts, max_distance=5, is_latlon=False).get_clusters()
        actual = [(c.centroid, len(c.points)) for c in actual]
        expected = [(c.geometry, c.n) for key, c in expected.iterrows()]
        assert len(actual) == len(expected)
        for pt in expected:
            assert pt in actual

    def test_cluster_vertical_points(self):
        pts = [
            Point(0, 2),
            Point(6, 2),
            Point(6, 2),
            Point(0.2, 2),
            Point(6.2, 2),
            Point(6.2, 2),
        ]
        expected = GeoDataFrame(
            pd.DataFrame(
                [
                    {"geometry": Point(0.1, 2), "n": 2},
                    {"geometry": Point(6.1, 2), "n": 4},
                ]
            )
        )
        actual = PointClusterer(pts, max_distance=5, is_latlon=False).get_clusters()
        actual = [(c.centroid, len(c.points)) for c in actual]
        expected = [(c.geometry, c.n) for key, c in expected.iterrows()]
        assert len(actual) == len(expected)
        print([str(c) for p in actual for c in p])
        for pt in expected:
            assert pt in actual



# -*- coding: utf-8 -*-

from movingpandas.trajectory import Trajectory
from movingpandas.trajectory_collection import TrajectoryCollection
from .test_trajectory import make_traj, Node
from movingpandas.trajectory_cleaner import IqrCleaner, OutlierCleaner
import pytest
import pandas as pd
from pyproj import CRS
from shapely.geometry import Point
from datetime import datetime
from geopandas import GeoDataFrame

CRS_METRIC = CRS.from_user_input(31256)
CRS_LATLON = CRS.from_user_input(4326)


class TestTrajectoryCleaner:
    def setup_method(self):
        df = pd.DataFrame(
            [
                [1, "A", Point(0, 0), datetime(2018, 1, 1, 12, 0, 0), 3, 10],
                [1, "A", Point(1, 1), datetime(2018, 1, 1, 12, 1, 0), 1, 14],
                [1, "A", Point(3, 1), datetime(2018, 1, 1, 12, 2, 0), 3, 11],
                [1, "A", Point(6, 0), datetime(2018, 1, 1, 12, 6, 0), -10, 20],
                [1, "A", Point(6, 6), datetime(2018, 1, 1, 14, 10, 0), 20, 200],
                [1, "A", Point(9, 9), datetime(2018, 1, 1, 14, 15, 0), 2, 20],
                [2, "A", Point(10, 10), datetime(2018, 1, 1, 12, 0, 0), 30, 20],
                [2, "A", Point(16, 10), datetime(2018, 1, 1, 12, 6, 0), 30, 30],
                [2, "A", Point(16, 12), datetime(2018, 1, 1, 12, 9, 0), 40, 30],
                [2, "A", Point(20, 20), datetime(2018, 1, 1, 12, 20, 0), 1000, 30],
                [2, "A", Point(16, 16), datetime(2018, 1, 2, 13, 10, 0), 20, -30],
                [2, "A", Point(190, 19), datetime(2018, 1, 2, 13, 15, 0), 20, 50],
            ],
            columns=["id", "obj", "geometry", "t", "val", "val2"],
        ).set_index("t")
        self.geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        self.collection = TrajectoryCollection(self.geo_df, "id", obj_id_col="obj")

    def test_outlier_cleaner_traj(self):
        traj = make_traj(
            [
                Node(),
                Node(1, 0.1, day=2, value=1),
                Node(2, 0.2, day=3, value=3),
                Node(3, 0, day=4, value=10),
                Node(3, 3, day=5, value=2),
                Node(4, 3, day=6, value=1),
                Node(5, 5, day=7, value=1),
            ]
        )
        result = IqrCleaner(traj).clean(columns={"value": 3})
        assert result == make_traj(
            [
                Node(),
                Node(1, 0.1, day=2, value=1),
                Node(2, 0.2, day=3, value=3),
                Node(3, 3, day=5, value=2),
                Node(4, 3, day=6, value=1),
                Node(5, 5, day=7, value=1),
            ]
        )

    def test_outlier_cleaner_collection(self):
        collection = IqrCleaner(self.collection).clean(columns={"val": 3, "val2": 3})
        assert len(collection) == 2
        wkt1 = collection.trajectories[0].to_linestring().wkt
        assert wkt1 == "LINESTRING (0 0, 1 1, 3 1, 9 9)"
        wkt2 = collection.trajectories[1].to_linestring().wkt
        assert wkt2 == "LINESTRING (10 10, 16 10, 16 12, 190 19)"

    def test_outlier_cleaner(self):
        df = pd.DataFrame(
            [
                [datetime(2013, 7, 1, 2, 4, 9), Point(-8.5829, 41.1451)],
                [datetime(2013, 7, 1, 2, 4, 24), Point(-8.5843, 41.1464)],
                [datetime(2013, 7, 1, 2, 4, 39), Point(-8.6108, 41.1458)],
                [datetime(2013, 7, 1, 2, 4, 54), Point(-8.6100, 41.1464)],
                [datetime(2013, 7, 1, 2, 5, 9), Point(-8.6090, 41.1468)],
                [datetime(2013, 7, 1, 2, 5, 24), Point(-8.6089, 41.1470)],
                [datetime(2013, 7, 1, 2, 5, 39), Point(-8.5860, 41.1487)],
                [datetime(2013, 7, 1, 2, 5, 54), Point(-8.5872, 41.1492)],
                [datetime(2013, 7, 1, 2, 6, 9), Point(-8.5882, 41.1489)],
            ],
            columns=["t", "geometry"],
        )
        gdf = GeoDataFrame(df, crs=CRS_LATLON)
        traj = Trajectory(gdf, traj_id=1, t="t")

        cleaned = OutlierCleaner(traj).clean()
        cleaned = OutlierCleaner(traj).clean(alpha=2)

        cleaned = OutlierCleaner(traj).clean(v_max=100, units=("km", "h"))
        expected = "LINESTRING (-8.5829 41.1451, -8.5843 41.1464, -8.586 41.1487, -8.5872 41.1492, -8.5882 41.1489)"  # noqa F401
        assert cleaned.to_linestring().wkt == expected

        # when traj is invalid after cleaning (e.g. <2 points),
        # must give warning & return original traj
        with pytest.warns(UserWarning):
            cleaned = OutlierCleaner(traj).clean(v_max=0.001)
            assert cleaned.is_valid()
            assert cleaned == traj



# -*- coding: utf-8 -*-

from pandas import DataFrame
from pandas.testing import assert_frame_equal
from geopandas import GeoDataFrame
from shapely.geometry import Point, LineString
from pyproj import CRS
from datetime import datetime, timedelta
from movingpandas.trajectory import Trajectory
from movingpandas.trajectory_collection import TrajectoryCollection
from movingpandas.trajectory_aggregator import (
    TrajectoryCollectionAggregator,
    PtsExtractor,
)

CRS_METRIC = CRS.from_user_input(31256)
CRS_LATLON = CRS.from_user_input(4326)


def assert_frame_not_equal(*args, **kwargs):
    # Taken from https://stackoverflow.com/a/38778401/6046019
    try:
        assert_frame_equal(*args, **kwargs)
    except AssertionError:
        # frames are not equal
        pass
    else:
        # frames are equal
        raise AssertionError


class TestPtsExtractor:
    def test_sparse_trajectory(self):
        df = DataFrame(
            [
                {"id": 1, "geometry": Point(0, 0), "t": datetime(2018, 1, 1, 12, 0, 0)},
                {"id": 1, "geometry": Point(6, 0), "t": datetime(2018, 1, 1, 12, 6, 0)},
                {"id": 1, "geometry": Point(6, 6), "t": datetime(2018, 1, 1, 14, 9, 0)},
            ]
        ).set_index("t")
        geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        traj = Trajectory(geo_df, "id")
        extractor = PtsExtractor(traj, 5, 0, min_stop_duration=timedelta(hours=12))
        actual = extractor.find_significant_points()
        expected = [Point(0, 0), Point(6, 0), Point(6, 6)]
        assert len(actual) == len(expected)
        for pt in expected:
            assert pt in actual

    def test_dense_trajectory(self):
        df = DataFrame(
            [
                {"id": 1, "geometry": Point(0, 0), "t": datetime(2018, 1, 1, 12, 0, 0)},
                {"id": 1, "geometry": Point(1, 0), "t": datetime(2018, 1, 1, 12, 1, 0)},
                {"id": 1, "geometry": Point(2, 0), "t": datetime(2018, 1, 1, 12, 2, 0)},
                {"id": 1, "geometry": Point(3, 0), "t": datetime(2018, 1, 1, 12, 3, 0)},
                {"id": 1, "geometry": Point(4, 0), "t": datetime(2018, 1, 1, 12, 4, 0)},
                {"id": 1, "geometry": Point(5, 0), "t": datetime(2018, 1, 1, 12, 5, 0)},
                {"id": 1, "geometry": Point(6, 0), "t": datetime(2018, 1, 1, 12, 6, 0)},
                {"id": 1, "geometry": Point(6, 6), "t": datetime(2018, 1, 1, 14, 9, 0)},
            ]
        ).set_index("t")
        geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        traj = Trajectory(geo_df, "id")
        extractor = PtsExtractor(traj, 4, 0, min_stop_duration=timedelta(hours=12))
        actual = extractor.find_significant_points()
        expected = [Point(0, 0), Point(4, 0), Point(6, 0), Point(6, 6)]
        assert len(actual) == len(expected)
        for pt in expected:
            assert pt in actual

    def test_no_stops_found(self):
        df = DataFrame(
            [
                {"id": 1, "geometry": Point(0, 0), "t": datetime(2018, 1, 1, 10, 0, 0)},
                {"id": 1, "geometry": Point(1, 0), "t": datetime(2018, 1, 1, 10, 1, 0)},
                {"id": 1, "geometry": Point(2, 0), "t": datetime(2018, 1, 1, 10, 2, 0)},
                {"id": 1, "geometry": Point(3, 0), "t": datetime(2018, 1, 1, 12, 3, 0)},
                {"id": 1, "geometry": Point(4, 0), "t": datetime(2018, 1, 1, 12, 4, 0)},
                {"id": 1, "geometry": Point(5, 0), "t": datetime(2018, 1, 1, 12, 5, 0)},
                {"id": 1, "geometry": Point(6, 0), "t": datetime(2018, 1, 1, 14, 6, 0)},
                {"id": 1, "geometry": Point(7, 0), "t": datetime(2018, 1, 1, 14, 7, 0)},
            ]
        ).set_index("t")
        geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        traj = Trajectory(geo_df, "id")
        extractor = PtsExtractor(traj, 10, 2, min_stop_duration=timedelta(hours=10))
        actual = extractor.find_significant_points()
        for pt in actual:
            print(pt)
        expected = [Point(0, 0), Point(7, 0)]
        assert len(actual) == len(expected)
        for pt in expected:
            assert pt in actual

    def test_stops_found(self):
        df = DataFrame(
            [
                {"id": 1, "geometry": Point(0, 0), "t": datetime(2018, 1, 1, 10, 0, 0)},
                {"id": 1, "geometry": Point(1, 0), "t": datetime(2018, 1, 1, 10, 1, 0)},
                {"id": 1, "geometry": Point(2, 0), "t": datetime(2018, 1, 1, 10, 2, 0)},
                {"id": 1, "geometry": Point(3, 0), "t": datetime(2018, 1, 1, 12, 3, 0)},
                {"id": 1, "geometry": Point(4, 0), "t": datetime(2018, 1, 1, 12, 4, 0)},
                {"id": 1, "geometry": Point(5, 0), "t": datetime(2018, 1, 1, 12, 5, 0)},
                {"id": 1, "geometry": Point(6, 0), "t": datetime(2018, 1, 1, 14, 6, 0)},
                {"id": 1, "geometry": Point(7, 0), "t": datetime(2018, 1, 1, 14, 7, 0)},
            ]
        ).set_index("t")
        geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        traj = Trajectory(geo_df, "id")
        extractor = PtsExtractor(traj, 10, 2, min_stop_duration=timedelta(hours=1))
        actual = extractor.find_significant_points()
        for pt in actual:
            print(pt)
        expected = [Point(0, 0), Point(2, 0), Point(5, 0), Point(7, 0)]
        assert len(actual) == len(expected)
        for pt in expected:
            assert pt in actual


class TestTrajectoryCollectionAggregator:
    def setup_method(self):
        pt_coords = [(0, 0), (6, 0), (6, 6), (0.2, 0.2), (6.2, 0.2), (6.2, 6.2)]
        self.pts = [Point(x, y) for (x, y) in pt_coords]
        lst = [
            [1, "A", Point(0, 0), datetime(2018, 1, 1, 12, 0, 0)],
            [1, "A", Point(6, 0), datetime(2018, 1, 1, 12, 6, 0)],
            [1, "A", Point(6, 6), datetime(2018, 1, 1, 14, 10, 0)],
            [2, "B", Point(0.2, 0.2), datetime(2019, 1, 1, 12, 0, 0)],
            [2, "B", Point(6.2, 0.2), datetime(2019, 1, 1, 12, 6, 0)],
            [2, "B", Point(6.2, 6.2), datetime(2019, 1, 1, 14, 10, 0)],
        ]
        self.df = DataFrame(lst, columns=["id", "obj", "geometry", "t"]).set_index("t")

    def create_metric(self, with_obj_id=False):
        self.geo_df = GeoDataFrame(self.df, crs=CRS_METRIC)
        if with_obj_id:
            self.collection = TrajectoryCollection(self.geo_df, "id", obj_id_col="obj")
        else:
            self.collection = TrajectoryCollection(self.geo_df, "id")

        self.trajectory_aggregator = TrajectoryCollectionAggregator(
            self.collection, 5, 0, min_stop_duration=timedelta(hours=12)
        )
        self.expected_clusters = GeoDataFrame(
            DataFrame(
                [[Point(0.1, 0.1), 2], [Point(6.1, 0.1), 2], [Point(6.1, 6.1), 2]],
                columns=["geometry", "n"],
            )
        )
        self.expected_flows = GeoDataFrame(
            DataFrame(
                [
                    [LineString([Point(0.1, 0.1), Point(6.1, 0.1)]), 2, 1],
                    [LineString([Point(6.1, 0.1), Point(6.1, 6.1)]), 2, 1],
                ],
                columns=["geometry", "weight", "obj_weight"],
            )
        )
        if with_obj_id:
            self.expected_flows["obj_weight"] = 2

    def create_latlon(self):
        self.geo_df_latlon = GeoDataFrame(self.df).set_crs(
            CRS_LATLON, allow_override=True
        )
        self.geo_df_latlon.crs = CRS_LATLON  # try to fix
        # https://travis-ci.com/github/anitagraser/movingpandas/builds/177149345
        self.collection_latlon = TrajectoryCollection(
            self.geo_df_latlon, "id", obj_id_col="obj"
        )
        self.trajectory_aggregator_latlon = TrajectoryCollectionAggregator(
            self.collection_latlon, 100, 0, min_stop_duration=timedelta(hours=12)
        )

    def test_get_significant_points_gdf_crs_metric(self):
        self.create_metric()
        sig_points = self.trajectory_aggregator.get_significant_points_gdf()
        assert sig_points.crs == CRS_METRIC
        actual = sig_points.geometry.tolist()
        expected = self.pts
        assert len(actual) == len(expected)
        for pt in expected:
            assert pt in actual

    def test_get_significant_points_gdf_crs_latlon(self):
        self.create_latlon()
        crs = self.trajectory_aggregator_latlon.get_significant_points_gdf().crs
        assert crs == CRS_LATLON

    def test_get_flows_gdf_crs_metric(self):
        self.create_metric()
        flows = self.trajectory_aggregator.get_flows_gdf()
        assert flows.crs == CRS_METRIC
        assert_frame_equal(flows, self.expected_flows)
        self.expected_flows["obj_weight"] = 99
        assert_frame_not_equal(flows, self.expected_flows)

    def test_get_flows_gdf_crs_metric_obj(self):
        self.create_metric(with_obj_id=True)
        flows = self.trajectory_aggregator.get_flows_gdf()
        assert flows.crs == CRS_METRIC
        assert_frame_equal(flows, self.expected_flows)
        self.expected_flows["obj_weight"] = 99
        assert_frame_not_equal(flows, self.expected_flows)

    def test_get_flows_gdf_crs_latlon(self):
        self.create_latlon()
        assert self.trajectory_aggregator_latlon.get_flows_gdf().crs == CRS_LATLON

    def test_get_clusters_gdf_crs_metric(self):
        self.create_metric()
        clusters = self.trajectory_aggregator.get_clusters_gdf()
        assert clusters.crs == CRS_METRIC
        print(clusters)
        actual = list(zip(*clusters[["geometry", "n"]]))
        expected = list(zip(*self.expected_clusters[["geometry", "n"]]))
        for pt in expected:
            assert pt in actual

    def test_get_clusters_gdf_crs_latlon(self):
        self.create_latlon()
        assert self.trajectory_aggregator_latlon.get_clusters_gdf().crs == CRS_LATLON



# -*- coding: utf-8 -*-

from geopy import distance
from math import hypot
from multiprocessing import Pool
from itertools import repeat
from geopandas import GeoDataFrame
from shapely.geometry import MultiPoint, Point
from .trajectory import Trajectory
from .trajectory_collection import TrajectoryCollection
from .geometry_utils import mrr_diagonal
from .trajectory_utils import convert_time_ranges_to_segments
from .spatiotemporal_utils import TRangeWithTrajId


class TrajectoryStopDetector:
    """
    Detects stops in a trajectory.
    A stop is detected if the movement stays within an area of specified size for
    at least the specified duration.
    """

    def __init__(self, traj, n_threads=1):
        """
        Create StopDetector

        Parameters
        ----------
        traj : Trajectory or TrajectoryCollection
        n_threads: number of threads to use for computation (default: 1)
        """
        self.traj = traj
        self.n_threads = n_threads

    def get_stop_time_ranges(self, max_diameter, min_duration):
        """
        Returns detected stop start and end times.

        Parameters
        ----------
        max_diameter : float
            Maximum diameter for stop detection. (Distances will be calculated
            in meters using geodesic distance calculations for geographic projections
            (e.g. EPSG:4326 WGS84) and in CRS units for all other projections)
        min_duration : datetime.timedelta
            Minimum stop duration

        Returns
        -------
        list
            TemporalRanges of detected stops
        """
        if isinstance(self.traj, Trajectory):
            return self._process_traj(self.traj, max_diameter, min_duration)
        elif isinstance(self.traj, TrajectoryCollection):
            trajs = self.traj.trajectories
            if self.n_threads > 1:
                return self._process_traj_collection_multithreaded(
                    trajs, max_diameter, min_duration
                )
            else:
                return self._process_traj_collection(trajs, max_diameter, min_duration)
        else:
            raise TypeError

    def _process_traj_collection(self, trajs, max_diameter, min_duration):
        results = []
        for traj in trajs:
            for time_range in self._process_traj(traj, max_diameter, min_duration):
                results.append(time_range)
        return results

    def _process_traj_collection_multithreaded(self, trajs, max_diameter, min_duration):
        from movingpandas.tools._multi_threading import split_list

        p = Pool(self.n_threads)
        data = split_list(trajs, self.n_threads)

        results = []
        for stops in p.starmap(
            self._process_traj_collection,
            zip(data, repeat(max_diameter), repeat(min_duration)),
        ):
            results.extend(stops)
        return results

    def _process_traj(self, traj, max_diameter, min_duration):
        detected_stops = []
        pts, xs, ys, ts = [], [], [], []
        minx, miny = float("inf"), float("inf")
        maxx, maxy = float("-inf"), float("-inf")
        geom = MultiPoint()
        is_stopped = False
        previously_stopped = False

        for t, pt in traj.df[traj.get_geom_col()].items():
            pts.append(pt)
            xs.append(pt.x)
            ys.append(pt.y)
            ts.append(t)
            first_to_keep = 0

            if not is_stopped:  # remove points to the specified min_duration
                while (
                    len(pts[first_to_keep:]) > 2
                    and ts[-1] - ts[first_to_keep] >= min_duration
                ):
                    first_to_keep += 1
                pts = pts[first_to_keep:]
                xs = xs[first_to_keep:]
                ys = ys[first_to_keep:]
                ts = ts[first_to_keep:]

            minx = min(xs)
            miny = min(ys)
            maxx = max(xs)
            maxy = max(ys)

            is_stopped = False
            if len(pts) > 1:
                if traj.is_latlon:
                    d = distance.distance((miny, minx), (maxy, maxx)).meters
                else:
                    d = hypot(maxx - minx, maxy - miny)
                if d < max_diameter * 1.5:
                    geom = MultiPoint(pts)
                    if mrr_diagonal(geom, traj.is_latlon) < max_diameter:
                        is_stopped = True

            if not is_stopped and previously_stopped and len(pts) > 1:
                segment_end = ts[-2]
                segment_begin = ts[0]
                if (
                    segment_end - segment_begin >= min_duration
                ):  # detected end of a stop
                    detected_stops.append(
                        TRangeWithTrajId(segment_begin, segment_end, traj.id)
                    )
                    pts = [pts[-1]]
                    xs = [xs[-1]]
                    ys = [ys[-1]]
                    ts = [ts[-1]]

            previously_stopped = is_stopped

        if is_stopped and ts[-1] - ts[0] >= min_duration:
            detected_stops.append(TRangeWithTrajId(ts[0], ts[-1], traj.id))

        return detected_stops

    def get_stop_segments(self, max_diameter, min_duration):
        """
        Returns detected stop trajectory segments.

        Parameters
        ----------
        max_diameter : float
            Maximum diameter for stop detection. (Distances will be calculated
            in meters using geodesic distance calculations for geographic projections
            (e.g. EPSG:4326 WGS84) and in CRS units for all other projections)
        min_duration : datetime.timedelta
            Minimum stop duration

        Returns
        -------
        TrajectoryCollection
            Trajectory segments

        Examples
        --------

        >>> detector = mpd.TrajectoryStopDetector(traj)
        >>> stops = detector.get_stop_segments(min_duration=timedelta(seconds=60),
                                               max_diameter=100)
        """
        stop_time_ranges = self.get_stop_time_ranges(max_diameter, min_duration)
        return TrajectoryCollection(
            convert_time_ranges_to_segments(self.traj, stop_time_ranges)
        )

    def get_stop_points(self, max_diameter, min_duration):
        """
        Returns detected stop location points.

        Parameters
        ----------
        max_diameter : float
            Maximum diameter for stop detection. (Distances will be calculated
            in meters using geodesic distance calculations for geographic projections
            (e.g. EPSG:4326 WGS84) and in CRS units for all other projections)
        min_duration : datetime.timedelta
            Minimum stop duration

        Returns
        -------
        geopandas.GeoDataFrame
            Stop locations as points with start and end time and stop duration
            in seconds

        Examples
        --------

        >>> detector = mpd.TrajectoryStopDetector(traj)
        >>> stops = detector.get_stop_points(min_duration=timedelta(seconds=60),
                                             max_diameter=100)
        """
        stop_time_ranges = self.get_stop_time_ranges(max_diameter, min_duration)
        stops = TrajectoryCollection(
            convert_time_ranges_to_segments(self.traj, stop_time_ranges)
        )

        stop_pts = GeoDataFrame(columns=["geometry"]).set_geometry("geometry")
        stop_pts["stop_id"] = [track.id for track in stops.trajectories]
        stop_pts = stop_pts.set_index("stop_id")

        for stop in stops:
            stop_pts.at[stop.id, "start_time"] = stop.get_start_time()
            stop_pts.at[stop.id, "end_time"] = stop.get_end_time()
            pt = Point(stop.df.geometry.x.median(), stop.df.geometry.y.median())
            stop_pts.at[stop.id, "geometry"] = pt
            stop_pts.at[stop.id, "traj_id"] = stop.parent.id

        if stops:
            stop_pts["duration_s"] = (
                stop_pts["end_time"] - stop_pts["start_time"]
            ).dt.total_seconds()
            stop_pts["traj_id"] = stop_pts["traj_id"].astype(type(stop.parent.id))

        return stop_pts



from shapely.geometry import Point
from datetime import datetime
from .geometry_utils import measure_distance


class TPoint:
    """
    Temporal point
    """

    def __init__(self, t, pt) -> None:
        self.t = t
        self.pt = pt


class TRange:
    """
    Temporal range
    """

    def __init__(self, t_0, t_n):
        self.t_0 = t_0
        self.t_n = t_n


class STRange:
    """
    Spatiotemporal range
    """

    def __init__(self, pt_0, pt_n, t_0, t_n):
        if isinstance(type(pt_0), Point):
            raise TypeError("Input pt_0 has to be a shapely.geometry.Point!")
        if isinstance(type(pt_n), Point):
            raise TypeError("Input pt_n has to be a shapely.geometry.Point!")
        if isinstance(type(t_0), datetime):
            raise TypeError("Input t_0 has to be a datetime.datetime!")
        if isinstance(type(t_n), datetime):
            raise TypeError("Input t_n has to be a datetime.datetime!")
        self.pt_0 = pt_0
        self.pt_n = pt_n
        self.t_0 = t_0
        self.t_n = t_n


class TRangeWithTrajId:
    def __init__(self, t_0, t_n, traj_id):
        self.t_0 = t_0
        self.t_n = t_n
        self.traj_id = traj_id

    def __str__(self):
        dt = self.t_n - self.t_0
        return f"Traj {self.traj_id}: {self.t_0} - {self.t_n} (duration: {dt})"


def get_speed(tpt0, tpt1, is_latlon, conversion):
    d = measure_distance(tpt0.pt, tpt1.pt, is_latlon)
    d = d * conversion.crs / conversion.distance
    v = d / (tpt1.t - tpt0.t).total_seconds() * conversion.time
    return v


def get_speed2(pt0, pt1, delta_t, is_latlon, conversion):
    d = measure_distance(pt0, pt1, is_latlon)
    d = d * conversion.crs / conversion.distance
    v = d / delta_t.total_seconds() * conversion.time
    return v



# -*- coding: utf-8 -*-

from copy import copy
from shapely.geometry import LineString, Point
import pandas as pd

from .trajectory import Trajectory
from .trajectory_collection import TrajectoryCollection
from .geometry_utils import measure_distance


class TrajectoryGeneralizer:
    """
    Generalizer base class
    """

    def __init__(self, traj):
        """
        Create TrajectoryGeneralizer

        Parameters
        ----------
        traj : Trajectory or TrajectoryCollection
        """
        self.traj = traj
        self.traj_col_name = traj.get_geom_col()

    def generalize(self, tolerance):
        """
        Generalize the input Trajectory/TrajectoryCollection.

        Parameters
        ----------
        tolerance : any type
            Tolerance threshold, differs by generalizer

        Returns
        -------
        Trajectory/TrajectoryCollection
            Generalized Trajectory or TrajectoryCollection
        """
        if isinstance(self.traj, Trajectory):
            return self._generalize_traj(self.traj, tolerance)
        elif isinstance(self.traj, TrajectoryCollection):
            return self._generalize_traj_collection(tolerance)
        else:
            raise TypeError

    def _generalize_traj_collection(self, tolerance):
        generalized = []
        for traj in self.traj:
            generalized.append(self._generalize_traj(traj, tolerance))
        result = copy(self.traj)
        result.trajectories = generalized
        return result

    def _generalize_traj(self, traj, tolerance):
        return traj


class MinDistanceGeneralizer(TrajectoryGeneralizer):
    """
    Generalizes based on distance.

    This generalization ensures that consecutive locations are at least a
    certain distance apart.

    Distance is calculated using CRS units, except if the CRS is geographic
    (e.g. EPSG:4326 WGS84) then distance is calculated in metres.

    tolerance : float
        Desired minimum distance between consecutive points

    Examples
    --------

    >>> mpd.MinDistanceGeneralizer(traj).generalize(tolerance=1.0)
    """

    def _generalize_traj(self, traj, tolerance):
        temp_df = traj.df.copy()
        prev_pt = temp_df.iloc[0][traj.get_geom_col()]
        keep_rows = [0]
        for i, pt in enumerate(temp_df[traj.get_geom_col()]):
            dist = measure_distance(pt, prev_pt, traj.is_latlon)
            if dist >= tolerance:
                keep_rows.append(i)
                prev_pt = pt

        keep_rows.append(len(traj.df) - 1)
        new_df = traj.df.iloc[keep_rows]
        new_traj = Trajectory(new_df, traj.id, traj_id_col=traj.get_traj_id_col())
        return new_traj


class MinTimeDeltaGeneralizer(TrajectoryGeneralizer):
    """
    Generalizes based on time.

    This generalization ensures that consecutive rows are at least a certain
    timedelta apart.

    tolerance : datetime.timedelta
        Desired minimum time difference between consecutive rows

    Examples
    --------

    >>> mpd.MinTimeDeltaGeneralizer(traj).generalize(tolerance=timedelta(minutes=10))
    """

    def _generalize_traj(self, traj, tolerance):
        temp_df = traj.df.copy()
        temp_df["t"] = temp_df.index
        prev_t = temp_df["t"].iat[0]
        keep_rows = [0]

        for i, (_, row) in enumerate(temp_df.iterrows()):
            t = row["t"]
            tdiff = t - prev_t
            if tdiff >= tolerance:
                keep_rows.append(i)
                prev_t = t

        keep_rows.append(len(traj.df) - 1)
        new_df = traj.df.iloc[keep_rows]
        new_traj = Trajectory(new_df, traj.id, traj_id_col=traj.get_traj_id_col())
        return new_traj


class MaxDistanceGeneralizer(TrajectoryGeneralizer):
    """
    Generalizes based on distance.

    Similar to Douglas-Peuker. Single-pass implementation that checks whether
    the provided distance threshold is exceed.

    tolerance : float
        Distance tolerance in trajectory CRS units

    Examples
    --------

    >>> mpd.MaxDistanceGeneralizer(traj).generalize(tolerance=1.0)
    """

    def _generalize_traj(self, traj, tolerance):
        prev_pt = None
        pts = []
        keep_rows = [0]
        i = 0

        for current_pt in traj.df[traj.get_geom_col()]:
            if prev_pt is None:
                prev_pt = current_pt
            else:
                line = LineString([prev_pt, current_pt])
                if any(line.distance(pt) > tolerance for pt in pts):
                    prev_pt = current_pt
                    pts.clear()
                    keep_rows.append(i)
                pts.append(current_pt)
                i += 1

        keep_rows.append(i)
        new_df = traj.df.iloc[keep_rows]
        new_traj = Trajectory(new_df, traj.id, traj_id_col=traj.get_traj_id_col())
        return new_traj


class DouglasPeuckerGeneralizer(TrajectoryGeneralizer):
    """
    Generalizes using Douglas-Peucker algorithm (as implemented in shapely/Geos).

    tolerance : float
        Distance tolerance in trajectory CRS units

    References
    ----------
    * Douglas, D., & Peucker, T. (1973). Algorithms for the reduction of the number
      of points required to represent a digitized line or its caricature.
      The Canadian Cartographer 10(2), 112–122. doi:10.3138/FM57-6770-U75U-7727.

    Examples
    --------

    >>> mpd.DouglasPeuckerGeneralizer(traj).generalize(tolerance=1.0)
    """

    def _generalize_traj(self, traj, tolerance):
        keep_rows = []
        simplified = (
            traj.to_linestring().simplify(tolerance, preserve_topology=False).coords
        )

        for i, current_pt in enumerate(traj.df[traj.get_geom_col()]):
            if current_pt.coords[0] in simplified:
                keep_rows.append(i)

        new_df = traj.df.iloc[keep_rows]
        new_traj = Trajectory(new_df, traj.id, traj_id_col=traj.get_traj_id_col())
        return new_traj


class TopDownTimeRatioGeneralizer(TrajectoryGeneralizer):
    """
    Generalizes using Top-Down Time Ratio algorithm proposed by Meratnia & de By (2004).

    This is a spatiotemporal trajectory generalization algorithm. Where Douglas-Peucker
    simply measures the spatial distance between points and original line geometry,
    Top-Down Time Ratio (TDTR) measures the distance between points and their
    spatiotemporal projection on the trajectory. These projections are calculated based
    on the ratio of travel times between the segment start and end times and the point
    time.

    tolerance : float
        Distance tolerance (distance returned by shapely Point.distance function)

    References
    ----------
    * Meratnia, N., & de By, R.A. (2004). Spatiotemporal compression techniques for
      moving point objects. In International Conference on Extending Database Technology
      (pp. 765-782). Springer, Berlin, Heidelberg.

    Examples
    --------

    >>> mpd.TopDownTimeRatioGeneralizer(traj).generalize(tolerance=1.0)
    """

    def _generalize_traj(self, traj, tolerance):
        generalized = self.td_tr(traj.df.copy(), tolerance)
        return Trajectory(generalized, traj.id, traj_id_col=traj.get_traj_id_col())

    def td_tr(self, df, tolerance):
        if len(df) <= 2:
            return df
        else:
            de = (
                df.index.max().to_pydatetime() - df.index.min().to_pydatetime()
            ).total_seconds()

            t0 = df.index.min().to_pydatetime()

            pt0 = df[self.traj_col_name].iloc[0]
            ptn = df[self.traj_col_name].iloc[-1]

            dx = ptn.x - pt0.x
            dy = ptn.y - pt0.y

            dists = df.apply(
                lambda rec: self._dist_from_calced(rec, t0, pt0, de, dx, dy),
                axis=1,
            )

            if dists.max() > tolerance:
                return pd.concat(
                    [
                        self.td_tr(
                            df.iloc[: df.index.get_loc(dists.idxmax()) + 1], tolerance
                        ),
                        self.td_tr(
                            df.iloc[df.index.get_loc(dists.idxmax()) :], tolerance
                        ),
                    ]
                )
            else:
                return df.iloc[[0, -1]]

    def _dist_from_calced(self, rec, start_t, start_geom, de, dx, dy):
        di = (rec.name - start_t).total_seconds()
        calced = Point(start_geom.x + dx * di / de, start_geom.y + dy * di / de)
        return rec[self.traj_col_name].distance(calced)



# -*- coding: utf-8 -*-

import pandas as pd
from geopandas import GeoDataFrame
from shapely.geometry import Point
from pyproj import CRS
from datetime import datetime
from movingpandas.trajectory_collection import TrajectoryCollection
from movingpandas.trajectory_plotter import _TrajectoryPlotter

CRS_METRIC = CRS.from_user_input(31256)
CRS_LATLON = CRS.from_user_input(4326)


class TestTrajectoryCollection:
    def setup_method(self):
        df = pd.DataFrame(
            [
                [1, "A", Point(0, 0), datetime(2018, 1, 1, 12, 0, 0), 9, "a"],
                [1, "A", Point(6, 0), datetime(2018, 1, 1, 12, 0, 6), 5, "b"],
                [1, "A", Point(10, 0), datetime(2018, 1, 1, 12, 0, 10), 2, "c"],
                [1, "A", Point(20, 0), datetime(2018, 1, 1, 12, 0, 15), 4, "d"],
                [2, "A", Point(10, 10), datetime(2018, 1, 1, 12, 0, 0), 10, "e"],
                [2, "A", Point(16, 10), datetime(2018, 1, 1, 12, 0, 6), 6, "f"],
                [2, "A", Point(20, 10), datetime(2018, 1, 1, 12, 0, 10), 7, "g"],
                [2, "A", Point(35, 10), datetime(2018, 1, 1, 12, 0, 15), 3, "h"],
            ],
            columns=["id", "obj", "geometry", "t", "val", "val2"],
        ).set_index("t")
        self.geo_df = GeoDataFrame(df, crs=CRS_METRIC)
        self.collection = TrajectoryCollection(self.geo_df, "id", obj_id_col="obj")

    def test_get_min_max_values(self):
        self.plotter = _TrajectoryPlotter(self.collection, column="val")
        min_value, max_value = self.plotter.get_min_max_values()
        assert min_value == 2
        assert max_value == 10



# -*- coding: utf-8 -*-

import importlib
import platform
import sys


def _get_sys_info():
    """System information

    Returns
    -------
    sys_info : dict
        system and Python version information
    """
    python = sys.version.replace("\n", " ")

    blob = [
        ("python", python),
        ("executable", sys.executable),
        ("machine", platform.platform()),
    ]

    return dict(blob)


def _get_C_info():
    """Information on system PROJ, GDAL, GEOS
    Returns
    -------
    c_info: dict
        system PROJ information
    """
    try:
        import pyproj

        proj_version = pyproj.proj_version_str
    except Exception:
        proj_version = None
    try:
        import pyproj

        proj_dir = pyproj.datadir.get_data_dir()
    except Exception:
        proj_dir = None

    blob = [
        ("PROJ", proj_version),
        ("PROJ data dir", proj_dir),
    ]

    return dict(blob)


def _get_deps_info():
    """Overview of the installed version of main dependencies

    Returns
    -------
    deps_info: dict
        version information on relevant Python libraries
    """
    deps = [
        "numpy",
        "geopandas",
        "geopy",
        "geoviews",
        "holoviews",
        "hvplot",
        "mapclassify",
        "matplotlib",
        "pandas",
        "pyproj",
        "shapely",
        "stonesoup",
    ]

    def get_version(module):
        return module.__version__

    deps_info = {}

    for modname in deps:
        try:
            if modname in sys.modules:
                mod = sys.modules[modname]
            else:
                mod = importlib.import_module(modname)
            ver = get_version(mod)
            deps_info[modname] = ver
        except Exception:
            deps_info[modname] = None

    return deps_info


def show_versions():
    """
    Print system information and installed module versions.
    """
    from movingpandas import __version__ as mpd_version

    sys_info = _get_sys_info()
    deps_info = _get_deps_info()
    proj_info = _get_C_info()

    maxlen = max(len(x) for x in deps_info)
    print(f"\nMovingPandas {mpd_version}")
    print("\nSYSTEM INFO")
    print("-----------")
    for k, stat in sys_info.items():
        print(f"{k: <{maxlen}}: {stat}")
    print("\nPROJ INFO")
    print("-----------")
    for k, stat in proj_info.items():
        print(f"{k: <{maxlen}}: {stat}")
    print("\nPYTHON DEPENDENCIES")
    print("-------------------")
    for k, stat in deps_info.items():
        print(f"{k: <{maxlen}}: {stat}")



# -*- coding: utf-8 -*-

from math import atan2, cos, degrees, pi, radians, sin, sqrt

import shapely
from geopy import distance
from geopy.distance import geodesic
from packaging.version import Version
from shapely.geometry import LineString, Point

from movingpandas.tools._warnings import deprecated

try:
    SHAPELY_GE_2 = Version(shapely.__version__) >= Version("2.0.0")
except TypeError:
    SHAPELY_GE_2 = False

R_EARTH = 6371000  # radius of earth in meters
C_EARTH = 2 * R_EARTH * pi  # circumference


def _is_point(input):
    if not isinstance(input, Point):
        raise TypeError(
            f"Only Points are supported as arguments, got {input} {type(input)}"
        )


@deprecated
def measure_distance_spherical(point1, point2):
    """
    Return spherical distance between two shapely Points as a float.
    """
    _is_point(point1)
    _is_point(point2)
    lon1 = float(point1.x)
    lon2 = float(point2.x)
    lat1 = float(point1.y)
    lat2 = float(point2.y)
    delta_lat = radians(lat2 - lat1)
    delta_lon = radians(lon2 - lon1)
    a = sin(delta_lat / 2) * sin(delta_lat / 2) + cos(radians(lat1)) * cos(
        radians(lat2)
    ) * sin(delta_lon / 2) * sin(delta_lon / 2)
    c = 2 * atan2(sqrt(a), sqrt(1 - a))
    dist = R_EARTH * c
    return dist


def measure_distance_euclidean(point1, point2):
    """
    Return euclidean distance between two shapely Points as float.
    """
    _is_point(point1)
    _is_point(point2)
    return point1.distance(point2)


def measure_distance_geodesic(point1, point2):
    """
    This function calculates the geodesic distance between two points
    as a float of (SI) unit meters.

    Parameters
    ----------
    point1 : shapely.geometry.Point
        point object containing latitude / longitude defining a starting location
    point2 : shapely.geometry.Point
        object containing latitude / longitude defining a termination location

    Returns
    -------
    dist : float
        geodesic distance (on a WGS84 ellipsoid) in meters
    """
    _is_point(point1)
    _is_point(point2)
    lon1 = float(point1.x)
    lon2 = float(point2.x)
    lat1 = float(point1.y)
    lat2 = float(point2.y)
    dist = distance.distance(
        (lat1, lon1), (lat2, lon2)
    ).meters  # uses geodesic dist and defaults to WGS84
    return dist


def measure_distance(point1, point2, geodesic=False, conversion=None):
    """
    Convenience function that returns either euclidean or geodesic distance
    between two points
    """
    if geodesic:
        d = measure_distance_geodesic(point1, point2)
    else:
        d = measure_distance_euclidean(point1, point2)
    if conversion:
        d = d * conversion.crs / conversion.distance
    return d


def measure_distance_line(linestring, other, conversion):
    """
    Returns the euclidean distance between a linestring and other geometry
    """
    d = linestring.distance(other)
    return d / conversion.distance


def measure_length(geoseries, spherical=False, conversion=None):
    """
    Returns the total distance between the points of the geoseries
    """
    pt_tuples = [(pt.y, pt.x) for pt in geoseries.tolist()]
    if spherical:
        length = geodesic(*pt_tuples).m
    else:  # The following distance will be in CRS units that might not be meters!
        length = LineString(pt_tuples).length

    if conversion:
        length = length / conversion.distance

    return length


def calculate_initial_compass_bearing(point1, point2):
    """
    Calculate the bearing between two points.

    The formulae used is the following:
        θ = atan2(sin(Δlong).cos(lat2),
                  cos(lat1).sin(lat2) − sin(lat1).cos(lat2).cos(Δlong))
    :Parameters:
      - `point1: shapely Point
      - `point2: shapely Point
    :Returns:
      The bearing in degrees
    :Returns Type:
      float
    """
    _is_point(point1)
    _is_point(point2)
    lat1 = radians(point1.y)
    lat2 = radians(point2.y)
    delta_lon = radians(point2.x - point1.x)
    x = sin(delta_lon) * cos(lat2)
    y = cos(lat1) * sin(lat2) - (sin(lat1) * cos(lat2) * cos(delta_lon))
    initial_bearing = atan2(x, y)
    # Now we have the initial bearing but math.atan2 return values
    # from -180° to + 180° which is not what we want for a compass bearing
    # The solution is to normalize the initial bearing as shown below
    initial_bearing = degrees(initial_bearing)
    compass_bearing = (initial_bearing + 360) % 360

    return compass_bearing


def azimuth(point1, point2):
    """
    Calculates euclidean bearing of line between two points.
    """
    _is_point(point1)
    _is_point(point2)
    angle = atan2(point2.x - point1.x, point2.y - point1.y)
    azimuth = degrees(angle)
    if angle < 0:
        azimuth += 360
    return azimuth


def angular_difference(degrees1, degrees2):
    """
    Calculates the smaller angle between the provided bearings / headings.
    """
    diff = abs(degrees1 - degrees2)
    if diff > 180:
        diff = abs(diff - 360)
    return diff


def mrr_diagonal(geom, spherical=False):
    """
    Calculate the length of the diagonal of the minimum rotated rectangle of
    the input geometry.
    """
    if isinstance(geom, Point):
        return 0
    _geom = geom.geoms if SHAPELY_GE_2 else geom
    if len(_geom) == 2:
        return measure_distance(_geom[0], _geom[1], spherical)
    mrr = geom.minimum_rotated_rectangle
    try:  # usually mrr is a Polygon
        x, y = mrr.exterior.coords.xy
    except AttributeError:  # thrown if mrr is a LineString
        return measure_distance(Point(mrr.coords[0]), Point(mrr.coords[-1]), spherical)
    return measure_distance(Point(x[0], y[0]), Point(x[2], y[2]), spherical)


def point_gdf_to_linestring(df, geom_col_name):
    """
    Convert GeoDataFrame of Points to shapely LineString
    """
    if len(df) > 1:
        return LineString(df[geom_col_name].tolist())
    else:
        raise RuntimeError("DataFrame needs at least two points to make line!")



import warnings

from collections import namedtuple
from datetime import datetime


UNITS = namedtuple(
    "UNITS", "distance time time2 crs", defaults=(None, None, None, None)
)


DISTANCE_UNIT_LIST = [
    {"abbr": "km", "conv": 1000, "fullname": "Kilometer"},
    {"abbr": "m", "conv": 1, "fullname": "metre"},
    {"abbr": "dm", "conv": 0.1, "fullname": "Decimeter"},
    {"abbr": "cm", "conv": 0.01, "fullname": "Centimeter"},
    {"abbr": "mm", "conv": 0.001, "fullname": "Millimeter"},
    {"abbr": "nm", "conv": 1852.0, "fullname": "International Nautical Mile"},
    {"abbr": "inch", "conv": 0.0254, "fullname": "International Inch"},
    {"abbr": "ft", "conv": 0.3048, "fullname": "International Foot"},
    {"abbr": "yd", "conv": 0.9144, "fullname": "International Yard"},
    {"abbr": "mi", "conv": 1609.344, "fullname": "International Statute Mile"},
    {"abbr": "link", "conv": 0.201168, "fullname": "International Link"},
    {"abbr": "chain", "conv": 20.1168, "fullname": "International Chain"},
    {"abbr": "fathom", "conv": 1.8288, "fullname": "International Fathom"},
    {
        "abbr": "british_ft",
        "conv": 0.304799471538676,
        "fullname": "British foot (Sears 1922)",
        "epsg": 9041,
    },
    {
        "abbr": "british_yd",
        "conv": 0.914398414616029,
        "fullname": "British yard (Sears 1922)",
        "epsg": 9040,
    },
    {
        "abbr": "british_chain_sears",
        "conv": 20.11677651215526,
        "fullname": "British chain (Sears 1922)",
        "epsg": 9042,
    },
    {
        "abbr": "british_link_sears",
        "conv": 0.20116767651215526,
        "fullname": "British link (Sears 1922)",
        "epsg": 9043,
    },
    {"abbr": "sears_yd", "conv": 0.914398414616029, "fullname": "Yard (Sears)"},
    {"abbr": "link_sears", "conv": 0.20116767651215526, "fullname": "Link (Sears)"},
    {"abbr": "chain_sears", "conv": 20.11677651215526, "fullname": "Chain (Sears)"},
    {
        "abbr": "british_ft_sears_truncated",
        "conv": 0.914398,
        "fullname": "British foot (Sears 1922 truncated)",
        "epsg": 9300,
    },
    {
        "abbr": "british_chain_sears_truncated",
        "conv": 20.11676,
        "fullname": "British chain (Sears 1922 truncated)",
        "epsg": 9301,
    },
    {
        "abbr": "british_chain_benoit",
        "conv": 20.116782494375872,
        "fullname": "British chain (Benoit 1895 B)",
        "epsg": 9062,
    },
    {
        "abbr": "chain_benoit",
        "conv": 20.116782494375872,
        "fullname": "Chain (Benoit)",
        "epsg": 9062,
    },
    {
        "abbr": "link_benoit",
        "conv": 0.20116782494375872,
        "fullname": "Link (Benoit)",
        "epsg": 9063,
    },
    {
        "abbr": "clarke_yd",
        "conv": 0.9143917962,
        "fullname": "Clarke's yard",
        "epsg": 9037,
    },
    {
        "abbr": "clarke_ft",
        "conv": 0.3047972654,
        "fullname": "Clarke's Foot",
        "epsg": 9005,
    },
    {
        "abbr": "clarke_link",
        "conv": 0.201166195164,
        "fullname": "Clarke's link",
        "epsg": 9039,
    },
    {
        "abbr": "clarke_chain",
        "conv": 20.1166195164,
        "fullname": "Clarke's chain",
        "epsg": 9038,
    },
    {
        "abbr": "british_ft_1936",
        "conv": 0.3048007491,
        "fullname": "British foot (1936)",
        "epsg": 9095,
    },
    {
        "abbr": "gold_coast_ft",
        "conv": 0.3047997101815,
        "fullname": "Gold Coast foot",
        "epsg": 9094,
    },
    {"abbr": "rod", "conv": 0.1988387815, "fullname": "Rod"},
    {"abbr": "furlong", "conv": 201.168, "fullname": "Furlong"},
    {
        "abbr": "german_m",
        "conv": 1.0000135965,
        "fullname": "German legal metre",
        "epsg": 9031,
    },
    {"abbr": "survey_in", "conv": 0.0254000508001016, "fullname": "US survey inch"},
    {
        "abbr": "survey_ft",
        "conv": 0.3048006096012192,
        "fullname": "US survey foot",
        "epsg": 9003,
    },
    {"abbr": "survey_yd", "conv": 0.9144018288036575, "fullname": "US survey yard"},
    {
        "abbr": "survey_lk",
        "conv": 0.20116840233680463,
        "fullname": "US survey link",
        "epsg": 9034,
    },
    {
        "abbr": "survey_ch",
        "conv": 20.116840233680463,
        "fullname": "US survey chain",
        "epsg": 9033,
    },
    {
        "abbr": "survey_mi",
        "conv": 1609.3472186944373,
        "fullname": "US survey mile",
        "epsg": 9035,
    },
    {
        "abbr": "indian_yd",
        "conv": 0.914398530744441,
        "fullname": "Indian Yard",
        "epsg": 9084,
    },
    {
        "abbr": "indian_ft",
        "conv": 0.3047995104977167,
        "fullname": "Indian Foot",
        "epsg": 9080,
    },
    {
        "abbr": "indian_ft_1937",
        "conv": 0.30479841,
        "fullname": "Indian Foot",
        "epsg": 9081,
    },
    {
        "abbr": "indian_ft_1962",
        "conv": 0.3047996,
        "fullname": "Indian Foot",
        "epsg": 9082,
    },
    {
        "abbr": "indian_ft_1975",
        "conv": 0.3047995,
        "fullname": "Indian Foot",
        "epsg": 9083,
    },
    {
        "abbr": "deg",
        "conv": 1,
        "fullname": "degree",
        "epsg": 4326,
    },  # To allow geodesic conversions
]


TIME_UNIT_LIST = [
    {"abbr": "s", "conv": 1, "fullname": "seconds"},
    {"abbr": "min", "conv": 60, "fullname": "minutes"},
    {"abbr": "h", "conv": 3600, "fullname": "hours"},
    {"abbr": "d", "conv": 86400, "fullname": "days"},
    {"abbr": "a", "conv": 31557600, "fullname": "years"},
]


def to_unixtime(t):
    """
    Return float of total seconds since Unix time.
    """
    return (t - datetime(1970, 1, 1, 0, 0, 0)).total_seconds()


def get_conversion(units, crs_units):
    """
    Looks up unit conversions in the unit dictionaries
    If distance specified, lookup distance and crs conversions and check time
    If time specified, lookup time conversion and check if time2 specified
    If time2 specified, lookup time2 conversion
    Unit conversions default to 1 if not specified
    """
    d_conv, t_conv, t2_conv, crs_conv = 1, 1, 1, 1

    if isinstance(units, tuple):
        units = UNITS(*units)
    else:
        units = UNITS(units)

    if units.distance is not None:
        d_conv = next(
            (d["conv"] for d in DISTANCE_UNIT_LIST if d.get("abbr") == units.distance),
            None,
        )
        if d_conv is None:
            raise ValueError("Invalid distance units!")
        crs_conv = next(
            (d["conv"] for d in DISTANCE_UNIT_LIST if d.get("fullname") == crs_units),
            None,
        )
        if crs_conv is None:
            crs_conv = 1
            warnings.warn(
                "No valid CRS distance units. Computations will "
                "assume CRS distance units are meters",
                category=MissingCRSWarning,
            )
        if units.time is not None:
            t_conv = next(
                (t["conv"] for t in TIME_UNIT_LIST if t.get("abbr") == units.time), None
            )
            if t_conv is None:
                raise ValueError("Invalid time units!")
        if units.time2 is not None:
            t2_conv = next(
                (t["conv"] for t in TIME_UNIT_LIST if t.get("abbr") == units.time2),
                None,
            )
            if t2_conv is None:
                raise ValueError("Invalid second time units!")
    return UNITS(d_conv, t_conv, t2_conv, crs_conv)


class MissingCRSWarning(UserWarning, ValueError):
    pass



# -*- coding: utf-8 -*-

from pandas import concat
from copy import copy
from geopandas import GeoDataFrame
from .trajectory import (
    Trajectory,
    SPEED_COL_NAME,
    DIRECTION_COL_NAME,
    DISTANCE_COL_NAME,
    ACCELERATION_COL_NAME,
    ANGULAR_DIFFERENCE_COL_NAME,
    TIMEDELTA_COL_NAME,
)
from .trajectory_plotter import _TrajectoryPlotter
from .unit_utils import UNITS
from .io import gdf_to_mf_json


@staticmethod
def traj_to_tc(traj):
    return TrajectoryCollection([traj])


class TrajectoryCollection:
    def __init__(
        self,
        data,
        traj_id_col=None,
        obj_id_col=None,
        t=None,
        x=None,
        y=None,
        crs="epsg:4326",
        min_length=0,
        min_duration=None,
    ):
        """
        Create TrajectoryCollection from list of trajectories or GeoDataFrame

        Parameters
        ----------
        data : list[Trajectory] or GeoDataFrame or DataFrame
            List of Trajectory objects or a GeoDataFrame with trajectory IDs,
            point geometry column and timestamp index
        traj_id_col : string
            Name of the GeoDataFrame column containing trajectory IDs
        obj_id_col : string
            Name of the GeoDataFrame column containing moving object IDs
        t : string
            Name of the DataFrame column containing the timestamp
        x : string
            Name of the DataFrame column containing the x coordinate
        y : string
            Name of the DataFrame column containing the y coordinate
        crs : string
            CRS of the x/y coordinates
        min_length : numeric
            Desired minimum length of trajectories. Length is calculated using
            CRS units, except if the CRS is geographic (e.g. EPSG:4326 WGS84)
            then length is calculated in meters.
            (Shorter trajectories are discarded.)
        min_duration : timedelta
            Desired minimum duration of trajectories. (Shorter trajectories are
            discarded.)

        Examples
        --------
        >>> import geopandas as read_file
        >>> import movingpandas as mpd
        >>>
        >>> gdf = read_file('data.gpkg')
        >>> collection = mpd.TrajectoryCollection(gdf, 'trajectory_id', t='t')
        """
        self.min_length = min_length
        self.min_duration = min_duration
        self.t = t
        if type(data) == list:
            self.trajectories = [
                traj for traj in data if traj.get_length() >= min_length
            ]
            if min_duration:
                self.trajectories = [
                    traj
                    for traj in self.trajectories
                    if traj.get_duration() >= min_duration
                ]
        else:
            self.trajectories = self._df_to_trajectories(
                data, traj_id_col, obj_id_col, t, x, y, crs
            )

    def __len__(self):
        return len(self.trajectories)

    def __str__(self):
        return f"TrajectoryCollection with {self.__len__()} trajectories"

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        """
        Iterator for trajectories in this collection

        Examples
        --------
        >>>  for traj in trajectory_collection:
        >>>      print(traj)
        """
        for traj in self.trajectories:
            if len(traj.df) >= 2:
                yield traj
            else:
                raise ValueError(
                    f"Trajectory with length >= 2 expected: "
                    f"got length {len(traj.df)}"
                )

    def copy(self):
        """
        Return a copy of the trajectory collection.

        Returns
        -------
        TrajectoryCollection
        """
        trajectories = [traj.copy() for traj in self.trajectories]
        # NOTE: traj_id_col and obj_id_col not needed since trajectories are
        # already preprocessed on __init__().
        return TrajectoryCollection(trajectories, min_length=self.min_length)

    def drop(self, **kwargs):
        """
        Drop columns or rows from the trajectories' DataFrames

        Examples
        --------

        >>> tc.drop(columns=['abc','def'])
        """
        for traj in self.trajectories:
            traj.drop(**kwargs)

    def to_point_gdf(self):
        """
        Return the trajectories' points as GeoDataFrame.

        Returns
        -------
        GeoDataFrame
        """
        gdfs = [traj.to_point_gdf() for traj in self.trajectories]
        return concat(gdfs)

    def to_line_gdf(self, columns=None):
        """
        Return the trajectories' line segments as GeoDataFrame.

        Returns
        -------
        GeoDataFrame
        """
        gdfs = [traj.to_line_gdf(columns) for traj in self.trajectories]
        gdf = concat(gdfs)
        gdf.reset_index(drop=True, inplace=True)
        return gdf

    def to_traj_gdf(self, wkt=False, agg=False):
        """
        Return a GeoDataFrame with one row per Trajectory within the
        TrajectoryCollection

        Returns
        -------
        GeoDataFrame
        """
        gdfs = [traj.to_traj_gdf(wkt, agg) for traj in self.trajectories]
        gdf = concat(gdfs)
        gdf.reset_index(drop=True, inplace=True)
        return gdf

    def to_mf_json(self, datetime_to_str=True, temporal_columns=None):
        """
        Converts a TrajectoryCollection to a dictionary compatible with the Moving
        Features JSON (MF-JSON) specification.

        Examples
        --------

        >>> tc.to_mf_json()

        Returns:
            dict: The MF-JSON representation of the GeoDataFrame as a dictionary.
        """
        tmp = self.to_point_gdf()
        t = tmp.index.name
        mf_json = gdf_to_mf_json(
            tmp.reset_index(),
            self.get_traj_id_col(),
            t,
            datetime_to_str=datetime_to_str,
            temporal_columns=temporal_columns,
        )
        return mf_json

    def _df_to_trajectories(self, df, traj_id_col, obj_id_col, t, x, y, crs):
        trajectories = []
        for traj_id, values in df.groupby(traj_id_col):
            if len(values) < 2:
                continue
            if obj_id_col in values.columns:
                obj_id = values.iloc[0][obj_id_col]
            else:
                obj_id = None
            trajectory = Trajectory(
                values,
                traj_id,
                traj_id_col=traj_id_col,
                obj_id=obj_id,
                t=t,
                x=x,
                y=y,
                crs=crs,
            )
            if self.min_duration:
                if trajectory.get_duration() < self.min_duration:
                    continue
            if trajectory.df.geometry.count() < 2:
                continue
            if self.min_length > 0:
                if trajectory.get_length() < self.min_length:
                    continue
            if isinstance(df, GeoDataFrame):
                trajectory.crs = df.crs
            else:
                trajectory.crs = crs
            trajectories.append(trajectory)
        return trajectories

    def get_trajectory(self, traj_id):
        """
        Return the Trajectory with the requested ID

        Parameters
        ----------
        traj_id : any
            Trajectory ID

        Returns
        -------
        Trajectory
        """
        for traj in self:
            if traj.id == traj_id:
                return traj

    def get_trajectories(self, obj_id):
        """
        Return the trajectories of the requested moving object

        Parameters
        ----------
        obj_id : any
            Moving object ID

        Returns
        -------
        TrajectoryCollection
        """
        trajs = []
        for traj in self:
            if traj.obj_id == obj_id:
                trajs.append(traj)
        return TrajectoryCollection(trajs)

    def get_crs(self):
        """
        Return the CRS of the trajectories
        """
        return self.trajectories[0].get_crs()

    def is_latlon(self):
        """
        Return True if the trajectory CRS is geographic (e.g. EPSG:4326 WGS84)
        """
        return self.trajectories[0].is_latlon()

    def get_column_names(self):
        """
        Return the list of column names

        Returns
        -------
        list
        """
        return self.trajectories[0].df.columns

    def get_traj_id_col(self):
        """
        Return name of the trajectory ID column

        Returns
        -------
        string
        """
        return self.trajectories[0].get_traj_id_col()

    def get_geom_col(self):
        """
        Return name of the geometry column

        Returns
        -------
        string
        """
        return self.trajectories[0].get_geom_col()

    def get_speed_col(self):
        """
        Return name of the speed column

        Returns
        -------
        string
        """
        return self.trajectories[0].get_speed_col()

    def get_direction_col(self):
        """
        Return name of the direction column

        Returns
        -------
        string
        """
        return self.trajectories[0].get_direction_col()

    def get_locations_at(self, t, with_direction=False):
        """
        Returns GeoDataFrame with trajectory locations at the specified timestamp

        Parameters
        ----------
        t : datetime.datetime
            Timestamp to extract trajectory locations for

        Returns
        -------
        GeoDataFrame
            Trajectory locations at timestamp t
        """
        result = []

        if with_direction:
            direction_col = self.get_direction_col()
            direction_missing = direction_col not in self.get_column_names()

        for traj in self:
            if t == "start":
                tmp = traj.copy()
                if with_direction and direction_missing:
                    tmp.df = tmp.df.head(2)
                    tmp.add_direction(name=direction_col)
                x = tmp.get_row_at(tmp.get_start_time())
            elif t == "end":
                tmp = traj.copy()
                if with_direction and direction_missing:
                    tmp.df = tmp.df.tail(2)
                    tmp.add_direction(name=direction_col)
                x = tmp.get_row_at(tmp.get_end_time())
            else:
                if t < traj.get_start_time() or t > traj.get_end_time():
                    continue
                tmp = traj.copy()
                if with_direction and direction_missing:
                    tmp.add_direction(name=direction_col)
                x = tmp.get_row_at(t)
            result.append(x.to_frame().T)

        if result:
            df = concat(result)
            # Move temporal index to column t
            t = self.t or "t"
            df.reset_index(inplace=True)
            df.rename(columns={"index": t}, inplace=True)

            gdf = GeoDataFrame(df)
            gdf.set_crs(self.get_crs(), inplace=True)
        else:
            gdf = GeoDataFrame()
        return gdf

    def get_start_locations(self, with_direction=False):
        """
        Returns GeoDataFrame with trajectory start locations

        Returns
        -------
        GeoDataFrame
            Trajectory start locations
        """
        return self.get_locations_at("start", with_direction)

    def get_end_locations(self, with_direction=False):
        """
        Returns GeoDataFrame with trajectory end locations

        Returns
        -------
        GeoDataFrame
            Trajectory end locations
        """
        return self.get_locations_at("end", with_direction)

    def get_segments_between(self, t1, t2):
        """
        Return Trajectory segments between times t1 and t2.

        Parameters
        ----------
        t1 : datetime.datetime
            Start time for the segments
        t2 : datetime.datetime
            End time for the segments

        Returns
        -------
        TrajectoryCollection
            Extracted trajectory segments
        """
        segments = []
        for traj in self:
            if t1 > traj.get_end_time() or t2 < traj.get_start_time():
                continue
            try:
                seg = traj.get_segment_between(t1, t2)
            except ValueError:
                continue
            segments.append(seg)
        result = copy(self)
        result.trajectories = segments
        return result

    def get_intersecting(self, polygon):
        """
        Return trajectories that intersect the given polygon.

        Parameters
        ----------
        polygon : shapely.geometry.Polygon
            Polygon to intersect with
        Returns
        -------
        TrajectoryCollection
            Resulting intersecting trajectories
        """
        intersecting = []
        for traj in self:
            try:
                if traj.intersects(polygon):
                    intersecting.append(traj)
            except:  # noqa E722
                pass
        result = copy(self)
        result.trajectories = intersecting
        return result

    def intersection(self, feature, point_based=False):
        """
        Intersect trajectories with the given polygon feature.

        Feature attributes are appended to the trajectory's DataFrame.

        By default, the trajectory's line representation is clipped by the
        polygon. If pointbased=True, the trajectory's point representation is
        used instead, leading to shorter segments.

        Parameters
        ----------
        feature : shapely Feature
            Feature to intersect with
        point_based : bool
            Clipping method

        Returns
        -------
        TrajectoryCollection
            Intersecting trajectory segments
        """
        intersections = []
        for traj in self:
            try:
                for intersect in traj.intersection(feature, point_based):
                    if (
                        intersect.get_length() >= self.min_length
                    ):  # TODO also test min_duration
                        intersections.append(intersect)
            except:  # noqa E722
                pass
        result = copy(self)
        result.trajectories = intersections
        return result

    def clip(self, polygon, point_based=False):
        """
        Clip trajectories by the given polygon.

        Parameters
        ----------
        polygon : shapely.geometry.Polygon
            Polygon to clip with
        point_based : bool
            Clipping method

        Returns
        -------
        TrajectoryCollection
            Resulting clipped trajectory segments
        """
        clipped = []
        for traj in self:
            try:
                for intersect in traj.clip(polygon, point_based):
                    if (
                        intersect.get_length() >= self.min_length
                    ):  # TODO also test min_duration
                        clipped.append(intersect)
            except:  # noqa E722
                pass
        result = copy(self)
        result.trajectories = clipped
        return result

    def filter(self, property_name, property_values):
        """
        Filter trajectories by property

        A property is a value in the df that is constant for the whole trajectory.
        The filter only checks if the value on the first row equals the requested
        property value.

        Parameters
        ----------
        property_name : string
            Name of the DataFrame column containing the property
        property_values : list(any)
            Desired property values

        Returns
        -------
        TrajectoryCollection
            Trajectories that fulfill the filter criteria

        Examples
        --------
        >>> filtered = trajectory_collection.filter('object_type', ['TypeA', 'TypeB'])
        """
        filtered = []
        for traj in self:
            if type(property_values) == list:
                if traj.df.iloc[0][property_name] in property_values:
                    filtered.append(traj)
            else:
                if traj.df.iloc[0][property_name] == property_values:
                    filtered.append(traj)

        result = copy(self)
        result.trajectories = filtered
        return result

    def add_speed(
        self, overwrite=False, name=SPEED_COL_NAME, units=UNITS(), n_threads=1
    ):
        """
        Add speed column and values to the trajectories.

        Speed is calculated as CRS units per second, except if the CRS is geographic
        (e.g. EPSG:4326 WGS84) then speed is calculated in meters per second.

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing speed values (default: False)
        name : str
            Name of the speed column (default: speed)
        units : tuple(str)
            Units in which to calculate speed

            distance : str
                Abbreviation for the distance unit
                (default: CRS units, or metres if geographic)
            time : str
                Abbreviation for the time unit (default: seconds)

            For more info, check the list of supported units at
            https://movingpandas.org/units
        n_threads : int
            Number of threads to use for computation (default: 1)
        """
        if n_threads <= 1:
            self._add_speed(self.trajectories, name, units, overwrite)
        else:
            self._multithread(self._add_speed, n_threads, name, units, overwrite)
        return self

    def _add_speed(self, trajs, name, units, overwrite):
        for traj in trajs:
            traj.add_speed(name=name, units=units, overwrite=overwrite)
        return trajs

    def _multithread(self, fun, n_threads, name, units, overwrite):
        from multiprocessing import Pool
        from itertools import repeat
        from movingpandas.tools._multi_threading import split_list

        p = Pool(int(n_threads))
        data = split_list(self.trajectories, n_threads)
        self.trajectories = []
        args_iter = zip(data, repeat(name), repeat(units), repeat(overwrite))
        results = []
        for added in p.starmap(fun, args_iter):
            results.extend(added)
        self.trajectories = results
        return results

    def add_direction(self, overwrite=False, name=DIRECTION_COL_NAME, n_threads=1):
        """
        Add direction column and values to the trajectories.

        The direction is calculated between consecutive locations.
        Direction values are in degrees, starting North turning clockwise.

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing direction values (default: False)
        name : str
            Name of the direction column (default: "direction")
        n_threads : int
            Number of threads to use for computation (default: 1)
        """
        if n_threads <= 1:
            self._add_direction(self.trajectories, name, UNITS(), overwrite)
        else:
            self._multithread(
                self._add_direction,
                n_threads,
                name,
                UNITS(),
                overwrite,
            )
        return self

    def _add_direction(self, trajs, name, units, overwrite):
        for traj in trajs:
            traj.add_direction(overwrite=overwrite, name=name)
        return trajs

    def add_angular_difference(
        self, overwrite=False, name=ANGULAR_DIFFERENCE_COL_NAME, n_threads=1
    ):
        """
        Add angular difference to the trajectory's DataFrame.

        Angular difference is calculated as the absolute smaller angle
        between direction for points along the trajectory.
        Values are [0, 180.0]

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing angular difference values (default: False)
        name : str
            Name of the angular_difference column (default: "angular_difference")
        n_threads : int
            Number of threads to use for computation (default: 1)
        """
        if n_threads <= 1:
            self._add_angular_difference(self.trajectories, name, UNITS(), overwrite)
        else:
            self._multithread(
                self._add_angular_difference, n_threads, name, UNITS(), overwrite
            )
        return self

    def _add_angular_difference(self, trajs, name, units, overwrite):
        for traj in trajs:
            traj.add_angular_difference(overwrite=overwrite, name=name)
        return trajs

    def add_acceleration(
        self, overwrite=False, name=ACCELERATION_COL_NAME, units=UNITS(), n_threads=1
    ):
        """
        Add acceleration column and values to the trajectories.

        Acceleration is calculated as CRS units per second squared,
        except if the CRS is geographic (e.g. EPSG:4326 WGS84) then acceleration is
        calculated in meters per second squared.

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing acceleration values (default: False)
        name : str
            Name of the acceleration column (default: "acceleration")
        units : tuple(str)
            Units in which to calculate acceleration

            distance : str
                Abbreviation for the distance unit
                (default: CRS units, or metres if geographic)
            time : str
                Abbreviation for the time unit (default: seconds)
            time2 : str
                Abbreviation for the second time unit (default: seconds)

            For more info, check the list of supported units at
            https://movingpandas.org/units
        n_threads : int
            Number of threads to use for computation (default: 1)
        """
        if n_threads <= 1:
            self._add_acceleration(self.trajectories, name, units, overwrite)
        else:
            self._multithread(self._add_acceleration, n_threads, name, units, overwrite)
        return self

    def _add_acceleration(self, trajs, name, units, overwrite):
        for traj in trajs:
            traj.add_acceleration(overwrite=overwrite, name=name, units=units)
        return trajs

    def add_distance(
        self, overwrite=False, name=DISTANCE_COL_NAME, units=None, n_threads=1
    ):
        """
        Add distance column and values to the trajectories.

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing distance values (default: False)
        name : str
            Name of the distance column (default: "distance")
        units : str
            Units in which to calculate distance values (default: CRS units)
            For more info, check the list of supported units at
            https://movingpandas.org/units
        n_threads : int
            Number of threads to use for computation (default: 1)
        """
        if n_threads <= 1:
            self._add_distance(self.trajectories, name, units, overwrite)
        else:
            self._multithread(self._add_distance, n_threads, name, units, overwrite)
        return self

    def _add_distance(self, trajs, name, units, overwrite):
        for traj in trajs:
            traj.add_distance(overwrite=overwrite, name=name, units=units)
        return trajs

    def add_timedelta(self, overwrite=False, name=TIMEDELTA_COL_NAME, n_threads=1):
        """
        Add timedelta column and values to the trajectories.

        Timedelta is calculated as the time difference between the current
        and the previous row. Values are instances of datetime.timedelta.

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing timedelta values (default: False)
        name : str
            Name of the timedelta column (default: "timedelta")
        n_threads : int
            Number of threads to use for computation (default: 1)
        """
        if n_threads <= 1:
            self._add_timedelta(self.trajectories, name, UNITS(), overwrite)
        else:
            self._multithread(self._add_timedelta, n_threads, name, UNITS(), overwrite)
        return self

    def _add_timedelta(self, trajs, name, units, overwrite):
        for traj in trajs:
            traj.add_timedelta(overwrite=overwrite, name=name)
        return trajs

    def add_traj_id(self, overwrite=False):
        """
        Add trajectory id column and values to the trajectories.

        Parameters
        ----------
        overwrite : bool
            Whether to overwrite existing trajectory id values (default: False)
        """
        for traj in self:
            traj.add_traj_id(overwrite)
        return self

    def get_min(self, column):
        """
        Return minimum value in the provided DataFrame column over all trajectories

        Parameters
        ----------
        column : string
            Name of the DataFrame column

        Returns
        -------
        Sortable
            Minimum value
        """
        return min([traj.df[column].min() for traj in self])

    def get_max(self, column):
        """
        Return maximum value in the provided DataFrame column over all trajectories

        Parameters
        ----------
        column : string
            Name of the DataFrame column

        Returns
        -------
        Sortable
            Maximum value
        """
        return max([traj.df[column].max() for traj in self])

    def plot(self, *args, **kwargs):
        """
        Generate a plot.

        Parameters
        ----------
        args :
            These parameters will be passed to the TrajectoryPlotter
        kwargs :
            These parameters will be passed to the TrajectoryPlotter

        Examples
        --------
        Plot speed along trajectories (with legend and specified figure size):

        >>> trajectory_collection.plot(column='speed', legend=True, figsize=(9,5))

        Plot speed along trajectories (scaling the colormap to specific min/max values):

        >>> trajectory_collection.plot(column='speed', vmin=0, vmax=20)
        """
        return _TrajectoryPlotter(self, *args, **kwargs).plot()

    def explore(self, *args, **kwargs):
        """
        Generate a plot using GeoPandas explore (folium/leaflet.js)
        https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoDataFrame.explore.html

        Parameters
        ----------
        args :
            These parameters will be passed to GeoPandas explore
        kwargs :
            These parameters will be passed to GeoPandas explore

        Returns
        -------
        m : folium.folium.Map
            folium Map instance

        Examples
        --------
        Plot speed along trajectory (with legend and specified figure size):

        >>> trajectory_collection.explore(column='speed', vmax=20, tiles="CartoDB positron")
        """  # noqa: E501
        return _TrajectoryPlotter(self, *args, **kwargs).explore()

    def hvplot(self, *args, **kwargs):
        """
        Generate an interactive trajectory plot.

        Parameters
        ----------
        args :
            These parameters will be passed to the TrajectoryPlotter
        kwargs :
            These parameters will be passed to the TrajectoryPlotter

            To customize the plots, check the list of supported colormaps:
            https://holoviews.org/user_guide/Colormaps.html#available-colormaps

        Examples
        --------
        Plot speed along trajectories (with legend and specified figure size):

        >>> collection.hvplot(c='speed', line_width=7.0, width=700, height=400,
                              colorbar=True)

        Plot speed along trajectories (scaling the colormap to specific min/max values):

        >>> trajectory_collection.plot(column='speed', clim=(0,20))
        """  # noqa: E501
        return _TrajectoryPlotter(self, *args, **kwargs).hvplot()

    def hvplot_pts(self, *args, **kwargs):
        """
        Generate an interactive plot of trajectory points.

        Parameters
        ----------
        args :
            These parameters will be passed to the TrajectoryPlotter
        kwargs :
            These parameters will be passed to the TrajectoryPlotter

            To customize the plots, check the list of supported colormaps:
            https://holoviews.org/user_guide/Colormaps.html#available-colormaps

        Examples
        --------
        Plot points colored by speed (with legend and specified figure size):

        >>> collection.hvplot_pts(c='speed', width=700, height=400, colorbar=True)
        """  # noqa: E501
        return _TrajectoryPlotter(self, *args, **kwargs).hvplot_pts()


def _get_location_at(traj, t, columns=None):
    loc = {
        "t": t,
        "geometry": traj.get_position_at(t),
        "traj_id": traj.id,
        "obj_id": traj.obj_id,
    }
    if columns and columns != [None]:
        for column in columns:
            loc[column] = traj.df.iloc[traj.df.index.get_loc(t, method="nearest")][
                column
            ]
    return loc
